"""Methods that are shared among the inprocess algorithms."""
from __future__ import annotations

from typing import Optional, Tuple

from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC

from ethicml.utility import ClassifierType, KernelType

__all__ = ["settings_for_svm_lr"]


def settings_for_svm_lr(
    classifier: ClassifierType, C: Optional[float], kernel: Optional[KernelType]
) -> Tuple[float, Optional[KernelType]]:
    """If necessary get the default settings for the C and kernel parameter of SVM and LR."""
    if classifier is ClassifierType.gbt:
        return 1.0, None
    if C is None:
        if classifier is ClassifierType.lr:
            C = LogisticRegression().C
        elif classifier is ClassifierType.svm:
            C = SVC().C
        else:
            raise NotImplementedError(f'Unsupported classifier "{classifier}".')

    if kernel is None:
        kernel = KernelType[SVC().kernel] if classifier is ClassifierType.svm else None
    return C, kernel



"""Paper: "Achieving Equalized Odds by Resampling Sensitive Attributes," Y. Romano, S. Bates, and E. J. Candès, 2020.

https://github.com/yromano/fair_dummies
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import List
from typing_extensions import Literal, TypedDict

from ranzen import implements

from ethicml.algorithms.inprocess.in_algorithm import HyperParamType
from ethicml.algorithms.inprocess.in_subprocess import InAlgorithmSubprocess

__all__ = ["AdvDebiasing", "AdvDebArgs"]


class AdvDebArgs(TypedDict):
    """Args for the Agarwal implementation."""

    lr: float
    n_clf_epochs: int
    n_adv_epochs: int
    n_epoch_combined: int
    batch_size: int
    model_type: Literal["deep_model", "linear_model"]
    lambda_vec: float


@dataclass
class AdvDebiasing(InAlgorithmSubprocess):
    """Adversarial Debiasing Method."""

    lr = 0.5
    n_clf_epochs = 2
    n_adv_epochs = 2
    n_epoch_combined = 40
    batch_size = 32
    model_type: Literal["deep_model", "linear_model"] = "deep_model"
    lambda_vec = 0.999999

    @implements(InAlgorithmSubprocess)
    def _get_path_to_script(self) -> List[str]:
        return ["-m", "ethicml.implementations.adv_debiasing_method"]

    @implements(InAlgorithmSubprocess)
    def _get_flags(self) -> AdvDebArgs:
        return {
            "lr": self.lr,
            "n_clf_epochs": self.n_clf_epochs,
            "n_adv_epochs": self.n_adv_epochs,
            "n_epoch_combined": self.n_epoch_combined,
            "batch_size": self.batch_size,
            "model_type": self.model_type,
            "lambda_vec": self.lambda_vec,
        }

    @implements(InAlgorithmSubprocess)
    def get_hyperparameters(self) -> HyperParamType:
        return {
            "lr": self.lr,
            "n_clf_epochs": self.n_clf_epochs,
            "n_adv_epochs": self.n_adv_epochs,
            "n_epoch_combined": self.n_epoch_combined,
            "batch_size": self.batch_size,
            "model_type": self.model_type,
            "lambda_vec": self.lambda_vec,
        }

    @implements(InAlgorithmSubprocess)
    def get_name(self) -> str:
        return "Adversarial Debiasing"



"""Lookup tables / switch statements for project level objects."""

from typing import Callable, Dict, List

from .dataset import Dataset
from .tabular_data.adult import Adult
from .tabular_data.compas import Compas
from .tabular_data.credit import Credit
from .tabular_data.crime import Crime
from .tabular_data.german import German
from .tabular_data.health import Health
from .tabular_data.non_binary_toy import NonBinaryToy
from .tabular_data.sqf import Sqf
from .tabular_data.toy import Toy

__all__ = ["available_tabular", "get_dataset_obj_by_name"]


def _lookup_table() -> Dict[str, Callable[[], Dataset]]:
    return {
        Adult.__name__.lower(): Adult,
        Compas.__name__.lower(): Compas,
        Credit.__name__.lower(): Credit,
        Crime.__name__.lower(): Crime,
        German.__name__.lower(): German,
        NonBinaryToy.__name__.lower(): NonBinaryToy,
        Health.__name__.lower(): Health,
        Sqf.__name__.lower(): Sqf,
        Toy.__name__.lower(): Toy,
    }


def get_dataset_obj_by_name(name: str) -> Callable[[], Dataset]:
    """Given a dataset name, get the corresponding dataset object.

    :param name: Name of the dataset.
    """
    lookup = _lookup_table()
    lowercase_name = name.lower()
    if lowercase_name not in lookup:
        raise NotImplementedError("That dataset doesn't exist")

    return lookup[lowercase_name]


def available_tabular() -> List[str]:
    """List of tabular dataset names."""
    return list(_lookup_table())



"""Kamiran and Calders 2012."""
from dataclasses import dataclass
from typing import Any, Dict, Optional

import numpy as np
import pandas as pd
import sklearn
from ranzen import implements
from sklearn.linear_model import LogisticRegression

from ethicml.algorithms.inprocess.in_algorithm import HyperParamType, InAlgorithm
from ethicml.algorithms.inprocess.shared import settings_for_svm_lr
from ethicml.algorithms.inprocess.svm import KernelType, select_svm
from ethicml.utility import ClassifierType, DataTuple, Prediction, SoftPrediction, TestTuple

__all__ = ["Kamiran", "compute_instance_weights"]


VALID_MODELS = {ClassifierType.lr, ClassifierType.svm}


@dataclass
class Kamiran(InAlgorithm):
    """An implementation of the Reweighing method from `Kamiran and Calders 2012 <https://link.springer.com/article/10.1007/s10115-011-0463-8>`_.

    Each sample is assigned an instance-weight based on the joing probability of S and Y which is
    used during training of a classifier.

    :param classifier: The classifier to use.
    :param C: The C parameter for the classifier.
    :param kernel: The kernel to use for the classifier if SVM selected.
    """

    classifier: ClassifierType = ClassifierType.lr
    C: Optional[float] = None
    kernel: Optional[KernelType] = None

    def __post_init__(self) -> None:
        self.chosen_c, self.chosen_kernel = settings_for_svm_lr(
            self.classifier, self.C, self.kernel
        )
        self.group_weights: Optional[Dict[str, Any]] = None

    @implements(InAlgorithm)
    def get_hyperparameters(self) -> HyperParamType:
        _hyperparameters: Dict[str, Any] = {"C": self.C}
        if self.classifier is ClassifierType.svm:
            assert self.kernel is not None
            _hyperparameters["kernel"] = self.kernel
        return _hyperparameters

    @implements(InAlgorithm)
    def get_name(self) -> str:
        lr_params = f" C={self.chosen_c}" if self.classifier is ClassifierType.lr else ""
        svm_params = (
            f" C={self.C}, kernel={self.chosen_kernel}"
            if self.classifier is ClassifierType.svm
            else ""
        )
        return f"Kamiran & Calders {self.classifier}{lr_params}{svm_params}"

    @implements(InAlgorithm)
    def fit(self, train: DataTuple, seed: int = 888) -> InAlgorithm:
        self.clf = self._train(
            train, classifier=self.classifier, C=self.chosen_c, kernel=self.chosen_kernel, seed=seed
        )
        return self

    @implements(InAlgorithm)
    def predict(self, test: TestTuple) -> Prediction:
        return self._predict(model=self.clf, test=test)

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        clf = self._train(
            train, classifier=self.classifier, C=self.chosen_c, kernel=self.chosen_kernel, seed=seed
        )
        return self._predict(model=clf, test=test)

    def _train(
        self,
        train: DataTuple,
        classifier: ClassifierType,
        C: float,
        kernel: Optional[KernelType],
        seed: int,
    ) -> sklearn.linear_model._base.LinearModel:
        if classifier is ClassifierType.svm:
            assert kernel is not None
            model = select_svm(C=C, kernel=kernel, seed=seed)
        else:
            random_state = np.random.RandomState(seed=seed)
            model = LogisticRegression(
                solver="liblinear", random_state=random_state, max_iter=50_00, C=C
            )
        weights = compute_instance_weights(train)["instance weights"]
        model.fit(
            train.x,
            train.y.to_numpy().ravel(),
            sample_weight=weights,
        )
        weights = weights.value_counts().rename_axis('weight').reset_index(name='count')
        groups = (
            pd.concat([train.s, train.y], axis=1)
            .groupby([train.s.name, train.y.name])
            .size()
            .reset_index(name="count")
        )
        self.group_weights = pd.merge(weights, groups, on="count").T.to_dict()
        return model

    def _predict(
        self, model: sklearn.linear_model._base.LinearModel, test: TestTuple
    ) -> Prediction:
        return SoftPrediction((model.predict_proba(test.x)), info=self.get_hyperparameters())


def compute_instance_weights(
    train: DataTuple, balance_groups: bool = False, upweight: bool = False
) -> pd.DataFrame:
    """Compute weights for all samples.

    :param train: The training data.
    :param balance_groups: Whether to balance the groups. When False, the groups are balanced as in
        `Kamiran and Calders 2012 <https://link.springer.com/article/10.1007/s10115-011-0463-8>`_.
        When True, the groups are numerically balanced. (Default: False)
    :param upweight: If balance_groups is True, whether to upweight the groups, or to downweight
        them. Downweighting is done by multiplying the weights by the inverse of the group size and
        is more numerically stable for small group sizes. (Default: False)
    :returns: A dataframe with the instance weights for each sample in the training data.
    """
    num_samples = len(train.x)
    s_unique, inv_indices_s, counts_s = np.unique(train.s, return_inverse=True, return_counts=True)
    _, inv_indices_y, counts_y = np.unique(train.y, return_inverse=True, return_counts=True)
    group_ids = (inv_indices_y * len(s_unique) + inv_indices_s).squeeze()
    gi_unique, inv_indices_gi, counts_joint = np.unique(
        group_ids, return_inverse=True, return_counts=True
    )
    if balance_groups:
        # Upweight samples according to the cardinality of their intersectional group
        if upweight:
            group_weights = num_samples / counts_joint
        # Downweight samples according to the cardinality of their intersectional group
        # - this approach should be preferred due to being more numerically stable
        # (very small counts can lead to very large weighted loss values when upweighting)
        else:
            group_weights = 1 - (counts_joint / num_samples)
    else:
        counts_factorized = np.outer(counts_y, counts_s).flatten()
        group_weights = counts_factorized[gi_unique] / (num_samples * counts_joint)

    return pd.DataFrame(group_weights[inv_indices_gi], columns=["instance weights"])



"""For assessing Calder-Verwer metric, :math:`1-(P(Y=1|S=1)-P(Y=1|S!=1))`."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric
from .metric import MetricStaticName
from .per_sensitive_attribute import diff_per_sensitive_attribute, metric_per_sensitive_attribute
from .prob_pos import ProbPos

__all__ = ["AbsCV", "CV"]


@dataclass
class CV(CfmMetric):
    """Calder-Verwer."""

    _name: ClassVar[str] = "CV"
    apply_per_sensitive: ClassVar[bool] = False

    @implements(MetricStaticName)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        prob_pos = ProbPos(pos_class=self.pos_class, labels=self.labels)
        per_sens = metric_per_sensitive_attribute(prediction, actual, metric=prob_pos)
        diffs = diff_per_sensitive_attribute(per_sens)

        return 1 - list(diffs.values())[0]


@dataclass
class AbsCV(CV):
    """Absolute value of Calder-Verwer.

    This metric is supposed to make it easier to compare results.
    """

    _name: ClassVar[str] = "CV absolute"

    @implements(MetricStaticName)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        cv_score = super().score(prediction, actual)
        # the following is equivalent to 1 - abs(diff)
        if cv_score > 1:
            return 2 - cv_score
        return cv_score



"""Runs given metrics on given algorithms for given datasets."""
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Sequence, Union
from typing_extensions import Literal

import pandas as pd

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm
from ethicml.algorithms.preprocess.pre_algorithm import PreAlgorithm
from ethicml.data.dataset import Dataset
from ethicml.data.load import load_data
from ethicml.evaluators.parallelism import run_in_parallel
from ethicml.metrics.metric import Metric
from ethicml.metrics.per_sensitive_attribute import (
    MetricNotApplicable,
    diff_per_sensitive_attribute,
    metric_per_sensitive_attribute,
    ratio_per_sensitive_attribute,
)
from ethicml.preprocessing.scaling import ScalerType, scale_continuous
from ethicml.preprocessing.splits import DataSplitter, RandomSplit
from ethicml.utility.data_structures import (
    DataTuple,
    EvalTuple,
    HyperParamValue,
    Prediction,
    Results,
    ResultsAggregator,
    TrainValPair,
    make_results,
)

__all__ = ["evaluate_models", "run_metrics", "load_results"]


def get_sensitive_combinations(metrics: List[Metric], train: DataTuple) -> List[str]:
    """Get all possible combinations of sensitive attribute and metrics.

    :param metrics:
    :param train:
    """
    poss_values = [f"{train.s.name}_{unique}" for unique in train.s.unique()]
    return [f"{s}_{m.name}" for s in poss_values for m in metrics]


def per_sens_metrics_check(per_sens_metrics: Sequence[Metric]) -> None:
    """Check if the given metrics allow application per sensitive attribute.

    :param per_sens_metrics:
    """
    for metric in per_sens_metrics:
        if not metric.apply_per_sensitive:
            raise MetricNotApplicable(
                f"Metric {metric.name} is not applicable per sensitive "
                f"attribute, apply to whole dataset instead"
            )


def run_metrics(
    predictions: Prediction,
    actual: EvalTuple,
    metrics: Sequence[Metric] = (),
    per_sens_metrics: Sequence[Metric] = (),
    diffs_and_ratios: bool = True,
    use_sens_name: bool = True,
) -> Dict[str, HyperParamValue]:
    """Run all the given metrics on the given predictions and return the results.

    :param predictions: DataFrame with predictions
    :param actual: EvalTuple with the labels
    :param metrics: list of metrics (Default: ())
    :param per_sens_metrics: list of metrics that are computed per sensitive attribute (Default: ())
    :param diffs_and_ratios: if True, compute diffs and ratios per sensitive attribute (Default: True)
    :param use_sens_name: if True, use the name of the senisitive variable in the returned results.
                        If False, refer to the sensitive varibale as `S`. (Default: True)
    """
    result: Dict[str, HyperParamValue] = {}
    if predictions.hard.isna().any(axis=None):  # type: ignore[arg-type]
        return {"algorithm_failed": 1.0}
    for metric in metrics:
        result[metric.name] = metric.score(predictions, actual)

    for metric in per_sens_metrics:
        per_sens = metric_per_sensitive_attribute(predictions, actual, metric, use_sens_name)
        if diffs_and_ratios:
            diff_per_sens = diff_per_sensitive_attribute(per_sens)
            ratio_per_sens = ratio_per_sensitive_attribute(per_sens)
            per_sens.update(diff_per_sens)
            per_sens.update(ratio_per_sens)
        for key, value in per_sens.items():
            result[f"{metric.name}_{key}"] = value
    result.update(predictions.info)
    return result  # SUGGESTION: we could return a DataFrame here instead of a dictionary


def load_results(
    dataset_name: str,
    transform_name: str,
    topic: Optional[str] = None,
    outdir: Path = Path(".") / "results",
) -> Optional[Results]:
    """Load results from a CSV file that was created by `evaluate_models`.

    :param dataset_name: name of the dataset of the results
    :param transform_name: name of the transformation that was used for the results
    :param topic: (optional) topic string of the results (Default: None)
    :param outdir: directory where the results are stored (Default: Path(".") / "results")
    :returns: DataFrame if the file exists; None otherwise
    """
    csv_file = _result_path(outdir, dataset_name, transform_name, topic)
    if csv_file.is_file():
        return make_results(csv_file)
    return None


def _result_path(
    outdir: Path, dataset_name: str, transform_name: str, topic: Optional[str]
) -> Path:
    base_name: str = "" if topic is None else f"{topic}_"
    return outdir / f"{base_name}{dataset_name}_{transform_name}.csv"


def _delete_previous_results(
    outdir: Path, datasets: List[Dataset], transforms: Sequence[PreAlgorithm], topic: Optional[str]
) -> None:
    for dataset in datasets:
        transform_list = ["no_transform"]
        transform_list.extend(preprocess_model.name for preprocess_model in transforms)
        for transform_name in transform_list:
            path_to_file: Path = _result_path(outdir, dataset.name, transform_name, topic)
            if path_to_file.exists():
                path_to_file.unlink()


class _DataInfo(NamedTuple):
    test: DataTuple
    dataset_name: str
    transform_name: str
    split_info: Dict[str, float]
    scaler: str


def evaluate_models(
    datasets: List[Dataset],
    *,
    preprocess_models: Sequence[PreAlgorithm] = (),
    inprocess_models: Sequence[InAlgorithm] = (),
    metrics: Sequence[Metric] = (),
    per_sens_metrics: Sequence[Metric] = (),
    repeats: int = 1,
    test_mode: bool = False,
    delete_previous: bool = True,
    splitter: Optional[DataSplitter] = None,
    topic: Optional[str] = None,
    fair_pipeline: bool = True,
    num_jobs: Optional[int] = None,
    scaler: Optional[ScalerType] = None,
    repeat_on: Literal["data", "model", "both"] = "both",
) -> Results:
    """Evaluate all the given models for all the given datasets and compute all the given metrics.

    :param datasets: List of dataset objects.
    :param preprocess_models: List of preprocess model objects. (Default: ())
    :param inprocess_models: List of inprocess model objects. (Default: ())
    :param metrics: List of metric objects. (Default: ())
    :param per_sens_metrics: List of metric objects that will be evaluated per sensitive attribute.
        (Default: ())
    :param repeats: Number of repeats to perform for the experiments. (Default: 1)
    :param test_mode: If True, only use a small subset of the data so that the models run faster.
        (Default: False)
    :param delete_previous: True by default. If True, delete previous results in the directory.
    :param splitter: Custom train-test splitter. (Default: None)
    :param topic: A string that identifies the run; the string is prepended to the filename.
        (Default: None)
    :param fair_pipeline: if True, run fair inprocess algorithms on the output of preprocessing.
        (Default: True)
    :param num_jobs: Number of parallel jobs; if None, the number of CPUs is used. (Default: None)
    :param scaler: Sklearn-style scaler to be used on the continuous features. (Default: None)
    :param repeat_on: Should the `data` or `model` seed be varied for each run? Or should they `both` be the same? (Default: "both")
    """
    per_sens_metrics_check(per_sens_metrics)
    if splitter is None:
        if repeat_on == "model":
            train_test_split: DataSplitter = RandomSplit(train_percentage=0.8, start_seed=None)
        else:
            train_test_split = RandomSplit(train_percentage=0.8, start_seed=0)
    else:
        train_test_split = splitter

    default_transform_name = "no_transform"

    outdir = Path(".") / "results"  # OS-independent way of saying './results'
    outdir.mkdir(exist_ok=True)

    if delete_previous:
        _delete_previous_results(outdir, datasets, preprocess_models, topic)

    all_results = ResultsAggregator()

    # ======================================= prepare data ========================================
    data_splits: List[TrainValPair] = []
    test_data: List[_DataInfo] = []  # contains the test set and other things needed for the metrics
    model_seeds: List[int] = []
    for dataset in datasets:
        for split_id in range(repeats):
            train: DataTuple
            test: DataTuple
            train, test, split_info = train_test_split(load_data(dataset), split_id=split_id)
            if scaler is not None:
                train, scaler_post = scale_continuous(dataset, train, scaler)
                test, _ = scale_continuous(dataset, test, scaler_post, fit=False)
            if test_mode:
                # take smaller subset of training data to speed up training
                train = train.get_n_samples()
            train = train.rename(f"{train.name} ({split_id})")
            data_splits.append(TrainValPair(train, test))
            model_seeds.append(0 if repeat_on == "data" else split_id)
            split_info.update({"split_id": split_id})
            test_data.append(
                _DataInfo(
                    test=test,
                    dataset_name=dataset.name,
                    transform_name=default_transform_name,
                    split_info=split_info,
                    scaler="None" if scaler is None else scaler.__class__.__name__,
                )
            )

        # load previous results
        csv_file = _result_path(outdir, dataset.name, default_transform_name, topic)
        all_results.append_from_csv(csv_file)

    # ============================= inprocess models on untransformed =============================
    all_predictions = run_in_parallel(
        algos=inprocess_models, data=data_splits, seeds=model_seeds, num_jobs=num_jobs
    )
    inprocess_untransformed = _gather_metrics(
        all_predictions, test_data, inprocess_models, metrics, per_sens_metrics, outdir, topic
    )
    all_results.append_df(inprocess_untransformed)

    # ===================================== preprocess models =====================================
    # run all preprocess models
    all_transformed = run_in_parallel(
        algos=preprocess_models, data=data_splits, seeds=model_seeds, num_jobs=num_jobs
    )

    # append the transformed data to `transformed_data`
    transformed_data: List[TrainValPair] = []
    transformed_test: List[_DataInfo] = []
    for transformed, pre_model in zip(all_transformed, preprocess_models):
        for (transf_train, transf_test), data_info in zip(transformed, test_data):
            transformed_data.append(TrainValPair(transf_train, transf_test))
            transformed_test.append(
                _DataInfo(
                    test=data_info.test,
                    dataset_name=data_info.dataset_name,
                    transform_name=pre_model.name,
                    split_info=data_info.split_info,
                    scaler="None" if scaler is None else scaler.__class__.__name__,
                )
            )

    # ============================= inprocess models on transformed ===============================
    if fair_pipeline:
        run_on_transformed = inprocess_models
    else:
        # if not fair pipeline, run only the non-fair models on the transformed data
        run_on_transformed = [model for model in inprocess_models if not model.is_fairness_algo]

    transf_preds = run_in_parallel(
        algos=run_on_transformed,
        data=transformed_data,
        seeds=[0] * len(transformed_data),
        num_jobs=num_jobs,
    )
    transf_results = _gather_metrics(
        transf_preds, transformed_test, run_on_transformed, metrics, per_sens_metrics, outdir, topic
    )
    all_results.append_df(transf_results)

    # ======================================== return all =========================================
    return all_results.results


def _gather_metrics(
    all_predictions: List[List[Prediction]],
    test_data: Sequence[_DataInfo],
    inprocess_models: Sequence[InAlgorithm],
    metrics: Sequence[Metric],
    per_sens_metrics: Sequence[Metric],
    outdir: Path,
    topic: Optional[str],
) -> Results:
    """Take a list of lists of predictions and compute all metrics.

    :param all_predictions:
    :param test_data:
    :param inprocess_models:
    :param metrics:
    :param per_sens_metrics:
    :param outdir:
    :param topic:
    """
    columns = ["dataset", "scaler", "transform", "model", "split_id"]

    # transpose `all_results` so that the order in the results dataframe is correct
    num_cols = len(all_predictions[0]) if all_predictions else 0
    all_predictions_t = [[row[i] for row in all_predictions] for i in range(num_cols)]

    all_results = ResultsAggregator()

    # compute metrics, collect them and write them to files
    for preds_for_dataset, data_info in zip(all_predictions_t, test_data):
        # ============================= handle results of one dataset =============================
        results_df = pd.DataFrame(columns=columns)  # create empty results dataframe
        predictions: Prediction
        for predictions, model in zip(preds_for_dataset, inprocess_models):
            # construct a row of the results dataframe
            hyperparameters: Dict[str, Union[str, float]] = {
                k: v if isinstance(v, (float, int)) else str(v)
                for k, v in model.get_hyperparameters().items()
            }

            seed = predictions.info["model_seed"]
            assert isinstance(seed, int)
            df_row: Dict[str, HyperParamValue] = {
                "dataset": data_info.dataset_name,
                "scaler": data_info.scaler,
                "transform": data_info.transform_name,
                "model": model.name,
                "model_seed": seed,
                **data_info.split_info,
                **hyperparameters,
            }
            df_row.update(run_metrics(predictions, data_info.test, metrics, per_sens_metrics))

            results_df = results_df.append(df_row, ignore_index=True, sort=False)

        # write results to CSV files and load previous results from the files if they already exist
        csv_file = _result_path(outdir, data_info.dataset_name, data_info.transform_name, topic)
        aggregator = ResultsAggregator(results_df)
        # put old results before new results -> prepend=True
        aggregator.append_from_csv(csv_file, prepend=True)
        aggregator.save_as_csv(csv_file)
        all_results.append_df(results_df)

    return all_results.results



"""Split into train and test data."""
import itertools
from abc import ABC, abstractmethod
from typing import Dict, Iterator, List, Optional, Tuple
from typing_extensions import Literal

import numpy as np
import pandas as pd
from numpy.random import RandomState
from ranzen import implements

from ethicml.utility import DataTuple
from ethicml.utility.data_helpers import shuffle_df

__all__ = [
    "BalancedTestSplit",
    "DataSplitter",
    "ProportionalSplit",
    "RandomSplit",
    "SequentialSplit",
    "train_test_split",
]


class DataSplitter(ABC):
    """Base class for classes that split data."""

    @abstractmethod
    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        """Split the given data.

        :param data: the data to split
        :param split_id: ID that makes this split unique; this affects for example the random seed
        :returns: A tuple consisting of two datatuples that are the result of the split and a
            dictionary that contains information about the split that is useful for logging.
        """


class SequentialSplit(DataSplitter):
    """Take the first N samples for train set and the rest for test set; no shuffle."""

    def __init__(self, train_percentage: float):
        self.train_percentage = train_percentage

    @implements(DataSplitter)
    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        del split_id
        train_len = round(self.train_percentage * len(data))

        train = data.apply_to_joined_df(lambda df: df.iloc[:train_len].reset_index(drop=True))
        train = train.rename(f"{data.name} - Train")

        test = data.apply_to_joined_df(lambda df: df.iloc[train_len:].reset_index(drop=True))
        test = test.rename(f"{data.name} - Test")

        assert len(train) + len(test) == len(data)
        return train, test, {}


def train_test_split(
    data: DataTuple,
    train_percentage: Optional[float] = 0.8,
    random_seed: int = 0,
    num_test_samples: Optional[int] = None,
) -> Tuple[DataTuple, DataTuple]:
    """Split a data tuple into two datatuple along the rows of the DataFrames.

    :param data: data tuple to split
    :param train_percentage: percentage for train split. Must be `None` if `num_test_samples` given. (Default: 0.8)
    :param random_seed: seed to make splitting reproducible (Default: 0)
    :param num_test_samples: Number of samples to make the test set. Must be `None` if `train_percentage` given. (Default: None)
    :returns: train split and test split
    """
    if train_percentage is None:
        assert num_test_samples is not None
    if num_test_samples is None:
        assert train_percentage is not None

    # ======================== concatenate the datatuple to one dataframe =========================
    # save the column names for later
    x_columns: pd.Index = data.x.columns
    s_column = data.s.name
    y_column = data.y.name
    assert isinstance(s_column, str) and isinstance(y_column, str)

    all_data: pd.DataFrame = data.data

    all_data = shuffle_df(all_data, random_state=1)

    # ============================== split the concatenated dataframe =============================
    # permute
    all_data = shuffle_df(all_data, random_state=random_seed)

    # split
    if train_percentage is not None:
        train_len = int(train_percentage * len(all_data))
    else:
        train_len = len(all_data) - num_test_samples
    all_data_train = all_data.iloc[:train_len]
    all_data_test = all_data.iloc[train_len:]

    assert isinstance(all_data_train, pd.DataFrame)
    assert isinstance(all_data_test, pd.DataFrame)

    all_data_train = all_data_train.reset_index(drop=True)
    all_data_test = all_data_test.reset_index(drop=True)

    # ================================== assemble train and test ==================================
    train: DataTuple = DataTuple(
        data=all_data_train,
        s_column=s_column,
        y_column=y_column,
        name=f"{data.name} - Train",
    )

    test: DataTuple = DataTuple(
        data=all_data_test,
        s_column=s_column,
        y_column=y_column,
        name=f"{data.name} - Test",
    )

    assert isinstance(train.x, pd.DataFrame)
    assert isinstance(test.x, pd.DataFrame)
    pd.testing.assert_index_equal(train.x.columns, x_columns)
    pd.testing.assert_index_equal(test.x.columns, x_columns)

    assert isinstance(train.s, pd.Series)
    assert isinstance(test.s, pd.Series)
    assert train.s.name == s_column
    assert test.s.name == s_column

    assert isinstance(train.y, pd.Series)
    assert isinstance(test.y, pd.Series)
    assert train.y.name == y_column
    assert test.y.name == y_column

    return train, test


class RandomSplit(DataSplitter):
    """Standard train test split.

    :param train_percentage: how much of the data to use for the train split
    :param start_seed: random seed for the first split
    """

    def __init__(self, train_percentage: float = 0.8, start_seed: Optional[int] = 0):
        super().__init__()
        self.start_seed = start_seed
        self.train_percentage = train_percentage

    def _get_seed(self, split_id: int) -> int:
        return self.start_seed + 2410 * split_id if self.start_seed is not None else 0

    @implements(DataSplitter)
    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        random_seed = self._get_seed(split_id)
        split_info: Dict[str, float] = {"seed": random_seed}
        return train_test_split(data, self.train_percentage, random_seed) + (split_info,)


def generate_proportional_split_indices(
    data: DataTuple, train_percentage: float, random_seed: int = 42
) -> Tuple[np.ndarray, np.ndarray]:
    """Generate the indices of the train and test splits using a proportional sampling scheme.

    :param data:
    :param train_percentage:
    :param random_seed:  (Default: 42)
    """
    # local random state that won't affect the global state
    random = RandomState(seed=random_seed)

    s_vals: List[int] = list(map(int, data.s.unique()))
    y_vals: List[int] = list(map(int, data.y.unique()))

    train_indices: List[np.ndarray] = []
    test_indices: List[np.ndarray] = []

    # iterate over all combinations of s and y
    for s, y in itertools.product(s_vals, y_vals):
        # find all indices for this group
        idx = ((data.s == s) & (data.y == y)).to_numpy().nonzero()[0]

        # shuffle and take subsets
        random.shuffle(idx)
        split_indices: int = round(len(idx) * train_percentage)
        # append index subsets to the list of train indices
        train_indices.append(idx[:split_indices])
        test_indices.append(idx[split_indices:])

    train_indices_ = np.concatenate(train_indices, axis=0)
    test_indices_ = np.concatenate(test_indices, axis=0)
    del train_indices
    del test_indices

    num_groups = len(s_vals) * len(y_vals)
    expected_train_len = round(len(data) * train_percentage)
    # assert that we (at least approximately) achieved the specified `train_percentage`
    # the maximum error occurs when all the group splits favor train or all favor test
    assert expected_train_len - num_groups <= len(train_indices_) <= expected_train_len + num_groups

    return train_indices_, test_indices_


class ProportionalSplit(RandomSplit):
    """Split into train and test while preserving the proportion of s and y."""

    @implements(DataSplitter)
    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        random_seed = self._get_seed(split_id)
        train_indices, test_indices = generate_proportional_split_indices(
            data, train_percentage=self.train_percentage, random_seed=random_seed
        )

        train: DataTuple = DataTuple(
            data=data.data.iloc[train_indices].reset_index(drop=True),
            s_column=data.s_column,
            y_column=data.y_column,
            name=f"{data.name} - Train",
        )

        test: DataTuple = DataTuple(
            data=data.data.iloc[test_indices].reset_index(drop=True),
            s_column=data.s_column,
            y_column=data.y_column,
            name=f"{data.name} - Test",
        )

        # assert that no data points got lost anywhere
        assert len(data) == len(train) + len(test)

        split_info: Dict[str, float] = {"seed": random_seed}

        return train, test, split_info


class BalancedTestSplit(RandomSplit):
    """Split data such that the test set is balanced and the training set is proportional.

    :param balance_type: how to do the balancing
    :param train_percentage: how much of the data to use for the train split
    :param start_seed: random seed for the first split
    """

    def __init__(
        self,
        balance_type: Literal["P(s|y)=0.5", "P(y|s)=0.5", "P(s,y)=0.25"] = "P(s|y)=0.5",
        train_percentage: float = 0.8,
        start_seed: int = 0,
    ):
        super().__init__(start_seed=start_seed, train_percentage=train_percentage)
        self.balance_type = balance_type

    @implements(DataSplitter)
    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        random_seed = self._get_seed(split_id)
        random = RandomState(seed=random_seed)

        s_vals: List[int] = list(map(int, data.s.unique()))
        y_vals: List[int] = list(map(int, data.y.unique()))

        train_indices: List[np.ndarray] = []
        test_indices: List[np.ndarray] = []

        num_test: Dict[Tuple[int, int], int] = {}
        # find out how many samples are available for the test set
        for s, y in itertools.product(s_vals, y_vals):
            # find all indices for this group
            idx = ((data.s == s) & (data.y == y)).to_numpy().nonzero()[0]
            # how many elements are in this "quadrant"
            quadrant_size = len(idx)
            # compute how many elements would be available for the test set
            num_test[(s, y)] = round(quadrant_size * (1 - self.train_percentage))

        # compute how much we should take for the test set to make it balanced
        if self.balance_type == "P(s|y)=0.5":
            minimize_over_s = {y: min(num_test[(s, y)] for s in s_vals) for y in y_vals}
            num_test_balanced = {(s, y): minimize_over_s[y] for s in s_vals for y in y_vals}
        elif self.balance_type == "P(y|s)=0.5":
            minimize_over_y = {s: min(num_test[(s, y)] for y in y_vals) for s in s_vals}
            num_test_balanced = {(s, y): minimize_over_y[s] for s in s_vals for y in y_vals}
        elif self.balance_type == "P(s,y)=0.25":
            smallest_quadrant = min(num_test[(s, y)] for s in s_vals for y in y_vals)
            num_test_balanced = {(s, y): smallest_quadrant for s in s_vals for y in y_vals}
        else:
            raise ValueError("Unknown balance_type")

        num_dropped = 0
        # iterate over all combinations of s and y
        for s, y in itertools.product(s_vals, y_vals):
            # find all indices for this group
            idx = ((data.s == s) & (data.y == y)).to_numpy().nonzero()[0]

            # shuffle and take subsets
            random.shuffle(idx)
            split_indices: int = round(len(idx) * self.train_percentage)
            # append index subsets to the list of train indices
            train_indices.append(idx[:split_indices])
            test_indices.append(idx[split_indices : (split_indices + num_test_balanced[(s, y)])])
            num_dropped += num_test[(s, y)] - num_test_balanced[(s, y)]

        train_idx = np.concatenate(train_indices, axis=0)
        test_idx = np.concatenate(test_indices, axis=0)

        train: DataTuple = DataTuple(
            data=data.data.iloc[train_idx].reset_index(drop=True),
            s_column=data.s_column,
            y_column=data.y_column,
            name=f"{data.name} - Train",
        )

        test: DataTuple = DataTuple(
            data=data.data.iloc[test_idx].reset_index(drop=True),
            s_column=data.s_column,
            y_column=data.y_column,
            name=f"{data.name} - Test",
        )

        unbalanced_test_len = round(len(data) * (1 - self.train_percentage))
        split_info = {
            "seed": random_seed,
            "percent_dropped": num_dropped / unbalanced_test_len,
            self.balance_type: 1,
        }

        return train, test, split_info


def fold_data(data: DataTuple, folds: int) -> Iterator[Tuple[DataTuple, DataTuple]]:
    """So much love to sklearn for making their source code open.

    :param data:
    :param folds:
    """
    indices: np.ndarray = np.arange(data.x.shape[0])

    fold_sizes: np.ndarray = np.full(folds, data.x.shape[0] // folds, dtype=np.int32)
    fold_sizes[: data.x.shape[0] % folds] += np.int32(1)

    current = 0
    for i, fold_size in enumerate(fold_sizes):
        start, stop = current, int(current + fold_size)
        val_inds: np.ndarray = indices[start:stop]
        train_inds = np.array([i for i in indices if i not in val_inds])  # probably inefficient

        train_x = data.x.iloc[train_inds].reset_index(drop=True)
        train_s = data.s.iloc[train_inds].reset_index(drop=True)  # type: ignore[call-overload]
        train_y = data.y.iloc[train_inds].reset_index(drop=True)  # type: ignore[call-overload]

        assert train_x.shape == (len(train_inds), data.x.shape[1])
        assert train_s.shape == (len(train_inds),)
        assert train_y.shape == (len(train_inds),)

        val_x = data.x.iloc[val_inds].reset_index(drop=True)
        val_s = data.s.iloc[val_inds].reset_index(drop=True)  # type: ignore[call-overload]
        val_y = data.y.iloc[val_inds].reset_index(drop=True)  # type: ignore[call-overload]

        assert val_x.shape == (len(val_inds), data.x.shape[1])
        assert val_s.shape == (len(val_inds),)
        assert val_y.shape == (len(val_inds),)

        yield DataTuple.from_df(
            x=train_x, s=train_s, y=train_y, name=f"{data.name} - train fold {i}"
        ), DataTuple.from_df(x=val_x, s=val_s, y=val_y, name=f"{data.name} - test fold {i}")

        current = stop



"""Useful methods that are used in some of the data objects."""
import functools
import warnings
from itertools import groupby
from typing import Any, Callable, Dict, List, Mapping, NamedTuple, Optional, Sequence, TypeVar
from typing_extensions import ParamSpec

__all__ = [
    "LabelGroup",
    "LabelSpec",
    "deprecated",
    "filter_features_by_prefixes",
    "get_discrete_features",
    "group_disc_feat_indices",
    "label_spec_to_feature_list",
    "simple_spec",
]

_P = ParamSpec("_P")
_T = TypeVar("_T")


def deprecated(func: Callable[_P, _T]) -> Callable[_P, _T]:
    """Decorate functions as deprecated.

    It will result in a warning being emitted when the function is used.
    """

    @functools.wraps(func)
    def new_func(*args: _P.args, **kwargs: _P.kwargs) -> Any:
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn(
            f"The {func.__name__} class is deprecated. "
            f"Use the function `{func.__name__.lower()}` instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func


class LabelGroup(NamedTuple):
    """Definition of a group of columns that should be interpreted as a single label."""

    columns: List[str]  # list of columns that belong to the group
    multiplier: int = 1  # multiplier assigned to this group (needed when combining groups)


LabelSpec = Mapping[str, LabelGroup]


def filter_features_by_prefixes(features: Sequence[str], prefixes: Sequence[str]) -> List[str]:
    """Filter the features by prefixes.

    :param features: list of features names
    :param prefixes: list of prefixes
    :returns: filtered feature names
    """
    res: List[str] = []
    for name in features:
        filtered = any(name.startswith(pref) for pref in prefixes)
        if not filtered:
            res.append(name)
    return res


def get_discrete_features(
    all_feats: List[str], feats_to_remove: List[str], cont_feats: List[str]
) -> List[str]:
    """Get a list of the discrete features in a dataset.

    :param all_feats: List of all features in the dataset.
    :param feats_to_remove: List of features that aren't used.
    :param cont_feats: List of continuous features in the dataset.
    :returns: List of features not marked as continuous or to be removed.
    """
    return [
        item
        for item in filter_features_by_prefixes(all_feats, feats_to_remove)
        if item not in cont_feats
    ]


def group_disc_feat_indices(disc_feat_names: List[str], prefix_sep: str = "_") -> List[slice]:
    """Group discrete features names according to the first segment of their name.

    Returns a list of their corresponding slices (assumes order is maintained).

    :param disc_feat_names: List of discrete feature names.
    :param prefix_sep: Separator between the prefix and the rest of the name. (Default: "_")
    :returns: List of slices.
    """

    def _first_segment(feature_name: str) -> str:
        return feature_name.split(prefix_sep)[0]

    group_iter = groupby(disc_feat_names, _first_segment)

    feature_slices: List[slice] = []
    start_idx = 0
    for _, group in group_iter:
        len_group = len(list(group))
        indices = slice(start_idx, start_idx + len_group)
        feature_slices.append(indices)
        start_idx += len_group

    return feature_slices


def flatten_dict(dictionary: Optional[Mapping[str, List[str]]]) -> List[str]:
    """Flatten a dictionary of lists by joining all lists to one big list."""
    if dictionary is None:
        return []
    return [x for inner_list in dictionary.values() for x in inner_list]


def reduce_feature_group(
    disc_feature_groups: Dict[str, List[str]],
    feature_group: str,
    to_keep: Sequence[str],
    remaining_feature_name: str,
) -> List[str]:
    """Drop all features in the given feature group except the ones in to_keep.

    :param disc_feature_groups: Dictionary of feature groups.
    :param feature_group: Name of the feature group that will be replaced by ``to_keep``.
    :param to_keep: List of features that will be kept in the feature group.
    :param remaining_feature_name: Name of the dummy feature that will be used to summarize the
        removed features.
    :returns: Flattened version of the modified dictionary of feature groups.
    """
    # first set the given feature group to just the value that we want to keep
    disc_feature_groups[feature_group] = list(to_keep)
    # then add a new dummy feature to the feature group. `load_data()` will create this for us
    disc_feature_groups[feature_group].append(f"{feature_group}{remaining_feature_name}")
    # then, regenerate the list of discrete features; just like it's done in the constructor
    return flatten_dict(disc_feature_groups)


def label_spec_to_feature_list(spec: LabelSpec) -> List[str]:
    """Extract all the feature column names from a dictionary of label specifications.

    :param spec: Dictionary of label specifications.
    """
    feature_list: List[str] = []
    for group in spec.values():
        feature_list += group.columns
    return feature_list


def simple_spec(label_defs: Mapping[str, Sequence[str]]) -> LabelSpec:
    """Create label specs for the most common case where columns contain 0s and 1s.

    :param label_defs: Mapping of label names to column names.
    :returns: Label specifications.
    """
    multiplier = 1
    label_spec = {}
    for name, columns in label_defs.items():
        label_spec[name] = LabelGroup(list(columns), multiplier=multiplier)
        num_columns = len(columns)
        # we assume here that the columns only contain 0s and 1s
        multiplier *= num_columns if num_columns > 1 else 2
    return label_spec



"""Class to describe features of the Communities and Crime dataset."""
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type

from ..dataset import LoadableDataset
from ..util import flatten_dict

__all__ = ["Crime", "CrimeSplits"]


class CrimeSplits(Enum):
    """Splits for the Crime dataset."""

    RACE_BINARY = "Race-Binary"
    CUSTOM = "Custom"


@dataclass
class Crime(LoadableDataset):
    """UCI Communities and Crime dataset."""

    split: CrimeSplits = CrimeSplits.RACE_BINARY

    Splits: ClassVar[Type[CrimeSplits]] = CrimeSplits
    """Shorthand for the Enum that defines the splits associated with this class."""

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "state": [
                "state_1",
                "state_2",
                "state_4",
                "state_5",
                "state_6",
                "state_8",
                "state_9",
                "state_10",
                "state_11",
                "state_12",
                "state_13",
                "state_16",
                "state_18",
                "state_19",
                "state_20",
                "state_21",
                "state_22",
                "state_23",
                "state_24",
                "state_25",
                "state_27",
                "state_28",
                "state_29",
                "state_32",
                "state_33",
                "state_34",
                "state_35",
                "state_36",
                "state_37",
                "state_38",
                "state_39",
                "state_40",
                "state_41",
                "state_42",
                "state_44",
                "state_45",
                "state_46",
                "state_47",
                "state_48",
                "state_49",
                "state_50",
                "state_51",
                "state_53",
                "state_54",
                "state_55",
                "state_56",
            ],
            ">0.06black": [">0.06black"],
            "high_crime": ["high_crime"],
        }
        discrete_features = flatten_dict(disc_feature_groups)

        continuous_features = [
            "population",
            "householdsize",
            "racepctblack",
            "racePctWhite",
            "racePctAsian",
            "racePctHisp",
            "agePct12t21",
            "agePct12t29",
            "agePct16t24",
            "agePct65up",
            "numbUrban",
            "pctUrban",
            "medIncome",
            "pctWWage",
            "pctWFarmSelf",
            "pctWInvInc",
            "pctWSocSec",
            "pctWPubAsst",
            "pctWRetire",
            "medFamInc",
            "perCapInc",
            "whitePerCap",
            "blackPerCap",
            "indianPerCap",
            "AsianPerCap",
            "HispPerCap",
            "NumUnderPov",
            "PctPopUnderPov",
            "PctLess9thGrade",
            "PctNotHSGrad",
            "PctBSorMore",
            "PctUnemployed",
            "PctEmploy",
            "PctEmplManu",
            "PctEmplProfServ",
            "PctOccupManu",
            "PctOccupMgmtProf",
            "MalePctDivorce",
            "MalePctNevMarr",
            "FemalePctDiv",
            "TotalPctDiv",
            "PersPerFam",
            "PctFam2Par",
            "PctKids2Par",
            "PctYoungKids2Par",
            "PctTeen2Par",
            "PctWorkMomYoungKids",
            "PctWorkMom",
            "NumIlleg",
            "PctIlleg",
            "NumImmig",
            "PctImmigRecent",
            "PctImmigRec5",
            "PctImmigRec8",
            "PctImmigRec10",
            "PctRecentImmig",
            "PctRecImmig5",
            "PctRecImmig8",
            "PctRecImmig10",
            "PctSpeakEnglOnly",
            "PctNotSpeakEnglWell",
            "PctLargHouseFam",
            "PctLargHouseOccup",
            "PersPerOccupHous",
            "PersPerOwnOccHous",
            "PersPerRentOccHous",
            "PctPersOwnOccup",
            "PctPersDenseHous",
            "PctHousLess3BR",
            "MedNumBR",
            "HousVacant",
            "PctHousOccup",
            "PctHousOwnOcc",
            "PctVacantBoarded",
            "PctVacMore6Mos",
            "MedYrHousBuilt",
            "PctHousNoPhone",
            "PctWOFullPlumb",
            "OwnOccLowQuart",
            "OwnOccMedVal",
            "OwnOccHiQuart",
            "RentLowQ",
            "RentMedian",
            "RentHighQ",
            "MedRent",
            "MedRentPctHousInc",
            "MedOwnCostPctInc",
            "MedOwnCostPctIncNoMtg",
            "NumInShelters",
            "NumStreet",
            "PctForeignBorn",
            "PctBornSameState",
            "PctSameHouse85",
            "PctSameCity85",
            "PctSameState85",
            "LandArea",
            "PopDens",
            "PctUsePubTrans",
            "LemasPctOfficDrugUn",
            "ViolentCrimesPerPop",
        ]

        features_to_remove = ["communityname", "fold"]
        features = discrete_features + continuous_features
        features = [feature for feature in features if feature not in features_to_remove]
        if self.split is CrimeSplits.RACE_BINARY:
            sens_attr_spec = ">0.06black"
            s_prefix = [">0.06black", "race", "white", "black", "indian", "Asian", "Hisp", "Other"]
            class_label_spec = "high_crime"
            class_label_prefix = ["high_crime", "Violent"]
        elif self.split is CrimeSplits.CUSTOM:
            sens_attr_spec = ""
            s_prefix = []
            class_label_spec = ""
            class_label_prefix = []
        else:
            raise NotImplementedError
        super().__init__(
            name=f"Crime {self.split.value}",
            num_samples=1994,
            filename_or_path="crime.csv",
            features=features,
            cont_features=continuous_features,
            s_prefix=s_prefix,
            sens_attr_spec=sens_attr_spec,
            class_label_prefix=class_label_prefix,
            class_label_spec=class_label_spec,
            discrete_only=self.discrete_only,
            discrete_feature_groups=disc_feature_groups,
        )



"""For assessing mean of logits."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction, SoftPrediction

from .metric import Metric, MetricStaticName

__all__ = ["ProbOutcome"]


@dataclass
class ProbOutcome(MetricStaticName):
    """Mean of logits."""

    _name: ClassVar[str] = "prob_outcome"
    pos_class: int = 1

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        return (
            (prediction.soft.sum(axis=0)[self.pos_class] / prediction.hard.size).item()
            if isinstance(prediction, SoftPrediction)
            else float("nan")
        )



"""Class to describe features of the Adult dataset.

Uses FolkTables: https://github.com/zykls/folktables
Paper: https://arxiv.org/abs/2108.04884
"""
# pylint: skip-file
# Pylint will go crazy as we're reimplementing the Dataset Init.
import contextlib
import os
from pathlib import Path
from typing import Generator, Iterable, List, Union

import numpy as np
import pandas as pd
from folktables import ACSDataSource, adult_filter, folktables, state_list
from ranzen import implements

from ethicml.utility import DataTuple

from ..dataset import Dataset
from ..util import flatten_dict, simple_spec

__all__ = ["AcsIncome", "AcsEmployment"]


@contextlib.contextmanager
def download_dir(root: Path) -> Generator[None, None, None]:
    curdir = os.getcwd()
    os.chdir(root.expanduser().resolve())
    try:
        yield
    finally:
        os.chdir(curdir)


class AcsBase(Dataset):
    split: str
    target: str

    def __init__(
        self,
        name: str,
        root: Union[str, Path],
        year: str,
        horizon: int,
        states: List[str],
        class_label_spec: str,
        class_label_prefix: List[str],
        discrete_only: bool = False,
        invert_s: bool = False,
    ):
        if isinstance(root, str):
            root = Path(root)
        self.root = root.expanduser().resolve()
        self.root.mkdir(exist_ok=True)

        self.year = year
        self.horizon = horizon
        self.survey = "person"
        self.states = states
        self._invert_s = invert_s

        assert all(state in state_list for state in states)

        state_string = "_".join(states)
        super().__init__(
            name=f"{name}_{year}_{horizon}_{state_string}_{self.split}",
            filename_or_path=self.root,
            features=[],
            cont_features=[],
            sens_attr_spec="",
            class_label_spec=class_label_spec,
            class_label_prefix=class_label_prefix,
            discrete_only=discrete_only,
            num_samples=0,
        )

    def _backend_load(
        self, dataframe: pd.DataFrame, *, labels_as_features: bool, ordered: bool
    ) -> DataTuple:
        # +++ BELOW HERE IS A COPY OF DATASET LOAD +++

        assert isinstance(dataframe, pd.DataFrame)

        feature_split = self.ordered_features if ordered else self.feature_split
        if labels_as_features:
            feature_split_x = feature_split["x"] + feature_split["s"] + feature_split["y"]
        else:
            feature_split_x = feature_split["x"]

        # =========================================================================================
        # Check whether we have to generate some complementary columns for binary features.
        # This happens when we have for example several races: race-asian-pac-islander etc, but we
        # want to have a an attribute called "race_other" that summarizes them all. Now the problem
        # is that this cannot be done before this point, because only here have we actually loaded
        # the data. So, we have to do it here, with all the information we can piece together.

        disc_feature_groups = self._discrete_feature_groups
        if disc_feature_groups is not None:
            for group in disc_feature_groups.values():
                if len(group) == 1:
                    continue
                for feature in group:
                    if feature in dataframe.columns:
                        continue  # nothing to do
                    missing_feature = feature

                    existing_features = [other for other in group if other in dataframe.columns]
                    assert len(existing_features) == len(group) - 1, "at most 1 feature missing"
                    # the dummy feature is the inverse of the existing feature
                    or_combination = dataframe[existing_features[0]] == 1
                    for other in existing_features[1:]:
                        or_combination |= dataframe[other] == 1
                    inverse: pd.Series = 1 - or_combination
                    dataframe = pd.concat(
                        [dataframe, inverse.to_frame(name=missing_feature)], axis="columns"
                    )

        # =========================================================================================
        x_data = dataframe[feature_split_x]
        s_data = dataframe[feature_split["s"]]
        y_data = dataframe[feature_split["y"]]

        if self._map_to_binary:
            s_data = (s_data + 1) // 2  # map from {-1, 1} to {0, 1}
            y_data = (y_data + 1) // 2  # map from {-1, 1} to {0, 1}

        if self._invert_s:
            assert s_data.nunique().values[0] == 2, "s must be binary"
            s_data = 1 - s_data

        # the following operations remove rows if a label group is not properly one-hot encoded
        s_data, s_mask = self._one_hot_encode_and_combine(s_data, label_type="s")
        if s_mask is not None:
            x_data = x_data.loc[s_mask].reset_index(drop=True)
            s_data = s_data.loc[s_mask].reset_index(drop=True)
            y_data = y_data.loc[s_mask].reset_index(drop=True)
        y_data, y_mask = self._one_hot_encode_and_combine(y_data, label_type="y")
        if y_mask is not None:
            x_data = x_data.loc[y_mask].reset_index(drop=True)
            s_data = s_data.loc[y_mask].reset_index(drop=True)
            y_data = y_data.loc[y_mask].reset_index(drop=True)

        return DataTuple.from_df(x=x_data, s=s_data, y=y_data, name=self.name)


class AcsIncome(AcsBase):
    """The ACS Income Dataset from EAAMO21/NeurIPS21 - Retiring Adult."""

    def __init__(
        self,
        root: Union[str, Path],
        year: str,
        horizon: int,
        states: List[str],
        split: str = "Sex",
        target_threshold: int = 50_000,
        discrete_only: bool = False,
        invert_s: bool = False,
    ):

        self.split = split
        self.target = "PINCP"
        self.target_threshold = target_threshold

        self.sens_lookup = {"Sex": "SEX", "Race": "RAC1P"}

        super().__init__(
            name="ACS_Income",
            root=root,
            class_label_spec="PINCP_1",
            class_label_prefix=["PINCP"],
            year=year,
            horizon=horizon,
            states=states,
            discrete_only=discrete_only,
            invert_s=invert_s,
        )

    @staticmethod
    def cat_lookup(key: str) -> Iterable:
        """Look up categories."""
        table = {
            "COW": range(1, 9),
            'MAR': range(1, 6),
            'RELP': range(18),
            'SEX': range(1, 3),
            'RAC1P': range(1, 10),
            'PINCP': range(2),
        }

        return table[key]

    @implements(Dataset)
    def load(self, ordered: bool = False, labels_as_features: bool = False) -> DataTuple:
        datasource = ACSDataSource(
            survey_year=self.year, horizon=f'{self.horizon}-Year', survey=self.survey
        )

        with download_dir(self.root):
            dataframe = datasource.get_data(states=self.states, download=True)

        disc_feats = [
            'COW',
            'MAR',
            'RELP',
            'SEX',
            'RAC1P',
            'PINCP',
        ]

        continuous_features = [
            'AGEP',
            'OCCP',
            'POBP',
            'SCHL',
            'WKHP',
        ]

        data_obj = folktables.BasicProblem(
            features=disc_feats + continuous_features,
            target=self.target,
            target_transform=lambda x: x > self.target_threshold,
            group=self.split,
            preprocess=adult_filter,
            postprocess=lambda x: np.nan_to_num(x, -1),
        )

        dataframe = data_obj._preprocess(dataframe)
        dataframe[data_obj.target] = dataframe[data_obj.target].apply(data_obj._target_transform)

        for feat in disc_feats:
            dataframe[feat] = (
                dataframe[feat]
                .astype('int')
                .astype(pd.CategoricalDtype(categories=self.cat_lookup(feat)))
            )

        dataframe[continuous_features] = dataframe[continuous_features].astype('int')

        dataframe = pd.get_dummies(dataframe[disc_feats + continuous_features])

        dataframe = dataframe.apply(data_obj._postprocess)

        cow_cols = [col for col in dataframe.columns if col.startswith('COW')]
        mar_cols = [col for col in dataframe.columns if col.startswith('MAR')]
        relp_cols = [col for col in dataframe.columns if col.startswith('RELP')]
        sex_cols = [col for col in dataframe.columns if col.startswith('SEX')]
        rac1p_cols = [col for col in dataframe.columns if col.startswith('RAC1P')]
        pincp_cols = [col for col in dataframe.columns if col.startswith('PINCP')]
        disc_feature_groups = {
            "COW": cow_cols,
            'MAR': mar_cols,
            'RELP': relp_cols,
            'SEX': sex_cols,
            'RAC1P': rac1p_cols,
            'PINCP': pincp_cols,
        }
        discrete_features = flatten_dict(disc_feature_groups)

        # set all the attributes manually that we couldn't set in the __init__
        self._features = discrete_features + continuous_features
        self._cont_features_unfiltered = continuous_features

        self._num_samples = dataframe.shape[0]

        self._discrete_feature_groups = disc_feature_groups

        if self.split == "Sex":
            self._sens_attr_spec = f"{self.sens_lookup[self.split]}_1"
            self._s_prefix = [self.sens_lookup[self.split]]
        elif self.split == "Race":
            self._sens_attr_spec = simple_spec(
                {self.sens_lookup[self.split]: disc_feature_groups[self.sens_lookup[self.split]]}
            )
            self._s_prefix = [self.sens_lookup[self.split]]
        elif self.split == "Sex-Race":
            self._sens_attr_spec = simple_spec(
                {
                    f"{self.sens_lookup['Sex']}": [f"{self.sens_lookup['Sex']}_1"],
                    f"{self.sens_lookup['Race']}": disc_feature_groups[
                        f"{self.sens_lookup['Race']}"
                    ],
                }
            )
            self._s_prefix = [self.sens_lookup["Sex"], self.sens_lookup["Race"]]
        else:
            raise NotImplementedError

        dataframe = dataframe.reset_index(drop=True)

        return self._backend_load(dataframe, labels_as_features=labels_as_features, ordered=ordered)


class AcsEmployment(AcsBase):
    """The ACS Employmment Dataset from EAAMO21/NeurIPS21 - Retiring Adult."""

    def __init__(
        self,
        root: Union[str, Path],
        year: str,
        horizon: int,
        states: List[str],
        split: str = "Sex",
        discrete_only: bool = False,
        invert_s: bool = False,
    ):

        self.split = split
        self.target = "ESR"

        self.sens_lookup = {"Sex": "SEX", "Race": "RAC1P"}

        super().__init__(
            name="ACS_Employment",
            root=root,
            class_label_spec="ESR_1",
            class_label_prefix=["ESR"],
            year=year,
            horizon=horizon,
            states=states,
            discrete_only=discrete_only,
            invert_s=invert_s,
        )

    @staticmethod
    def cat_lookup(key: str) -> Iterable:
        """Look up categories."""
        table = {
            'SCHL': range(1, 25),
            'MAR': range(1, 6),
            'SEX': range(1, 3),
            'DIS': range(1, 3),
            'ESP': range(1, 9),
            'MIG': range(1, 4),
            'CIT': range(1, 5),
            'MIL': range(1, 5),
            'ANC': (1, 2, 4, 8),
            'NATIVITY': range(1, 3),
            'RELP': range(18),
            'DEAR': range(1, 3),
            'DEYE': range(1, 3),
            'DREM': range(1, 3),
            'RAC1P': range(1, 10),
            'ESR': range(1, 7),
        }

        return table[key]

    @implements(Dataset)
    def load(self, ordered: bool = False, labels_as_features: bool = False) -> DataTuple:
        datasource = ACSDataSource(
            survey_year=self.year, horizon=f'{self.horizon}-Year', survey=self.survey
        )

        with download_dir(self.root):
            dataframe = datasource.get_data(states=self.states, download=True)

        disc_feats = [
            'SCHL',
            'MAR',
            'RELP',
            'DIS',
            'ESP',
            'CIT',
            'MIG',
            'MIL',
            'ANC',
            'NATIVITY',
            'DEAR',
            'DEYE',
            'DREM',
            'SEX',
            'RAC1P',
            'ESR',
        ]

        continuous_features = [
            'AGEP',
        ]

        data_obj = folktables.BasicProblem(
            features=disc_feats + continuous_features,
            target=self.target,
            target_transform=lambda x: x == 1,
            group=self.split,
            preprocess=lambda x: x,
            postprocess=lambda x: np.nan_to_num(x, -1),
        )

        dataframe = data_obj._preprocess(dataframe)
        dataframe[data_obj.target] = dataframe[data_obj.target].apply(data_obj._target_transform)
        dataframe = dataframe.apply(data_obj._postprocess)

        for feat in disc_feats:
            dataframe[feat] = (
                dataframe[feat]
                .astype('int')
                .astype(pd.CategoricalDtype(categories=self.cat_lookup(feat)))
            )

        dataframe[continuous_features] = dataframe[continuous_features].astype('int')

        dataframe = pd.get_dummies(dataframe[disc_feats + continuous_features])

        dataframe = dataframe.apply(data_obj._postprocess)

        schl_cols = [col for col in dataframe.columns if col.startswith('SCHL')]
        mar_cols = [col for col in dataframe.columns if col.startswith('MAR')]
        relp_cols = [col for col in dataframe.columns if col.startswith('RELP')]
        dis_cols = [col for col in dataframe.columns if col.startswith('DIS')]
        esp_cols = [col for col in dataframe.columns if col.startswith('ESP')]
        cit_cols = [col for col in dataframe.columns if col.startswith('CIT')]
        mig_cols = [col for col in dataframe.columns if col.startswith('MIG')]
        mil_cols = [col for col in dataframe.columns if col.startswith('MIL')]
        anc_cols = [col for col in dataframe.columns if col.startswith('ANC')]
        nativity_cols = [col for col in dataframe.columns if col.startswith('NATIVITY')]
        dear_cols = [col for col in dataframe.columns if col.startswith('DEAR')]
        deye_cols = [col for col in dataframe.columns if col.startswith('DEYE')]
        drem_cols = [col for col in dataframe.columns if col.startswith('DREM')]
        sex_cols = [col for col in dataframe.columns if col.startswith('SEX')]
        rac1p_cols = [col for col in dataframe.columns if col.startswith('RAC1P')]
        esr_cols = [col for col in dataframe.columns if col.startswith('ESR')]
        disc_feature_groups = {
            "SCHL": schl_cols,
            'MAR': mar_cols,
            'RELP': relp_cols,
            'DIS': dis_cols,
            'ESP': esp_cols,
            'CIT': cit_cols,
            'MIG': mig_cols,
            'MIL': mil_cols,
            'ANC': anc_cols,
            'NATIVITY': nativity_cols,
            'DEAR': dear_cols,
            'DEYE': deye_cols,
            'DREM': drem_cols,
            'SEX': sex_cols,
            'RAC1P': rac1p_cols,
            'ESR': esr_cols,
        }
        discrete_features = flatten_dict(disc_feature_groups)

        # set all the attributes manually that we couldn't set in the __init__
        self._features = discrete_features + continuous_features
        self._cont_features_unfiltered = continuous_features

        self._num_samples = dataframe.shape[0]

        self._discrete_feature_groups = disc_feature_groups

        if self.split == "Sex":
            self._sens_attr_spec = f"{self.sens_lookup[self.split]}_1"
            self._s_prefix = [self.sens_lookup[self.split]]
        elif self.split == "Race":
            self._sens_attr_spec = simple_spec(
                {self.sens_lookup[self.split]: disc_feature_groups[self.sens_lookup[self.split]]}
            )
            self._s_prefix = [self.sens_lookup[self.split]]
        elif self.split == "Sex-Race":
            self._sens_attr_spec = simple_spec(
                {
                    f"{self.sens_lookup['Sex']}": [f"{self.sens_lookup['Sex']}_1"],
                    f"{self.sens_lookup['Race']}": disc_feature_groups[
                        f"{self.sens_lookup['Race']}"
                    ],
                }
            )
            self._s_prefix = [self.sens_lookup["Sex"], self.sens_lookup["Race"]]
        else:
            raise NotImplementedError

        dataframe = dataframe.reset_index(drop=True)

        return self._backend_load(dataframe, labels_as_features=labels_as_features, ordered=ordered)



"""For assessing Theil Index metric: based on aif 360 implementation.

https://github.com/IBM/AIF360/blob/b5f4591c7c7374b9c88706827080a931209e7a24/aif360/metrics/classification_metric.py
in turn based on the paper https://arxiv.org/abs/1807.00787
"""
from dataclasses import dataclass
from typing import ClassVar

import numpy as np
from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

from .metric import MetricStaticName

__all__ = ["Theil"]


@dataclass
class Theil(MetricStaticName):
    """Theil Index."""

    _name: ClassVar[str] = "Theil_Index"

    @implements(MetricStaticName)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        y_true_df = actual.y
        y_pos_label = y_true_df.max()

        y_pred = prediction.hard.to_numpy().ravel()
        y_true = y_true_df.to_numpy().ravel()
        y_pred = (y_pred == y_pos_label).astype(np.float64)
        y_true = (y_true == y_pos_label).astype(np.float64)
        var_b = 1 + y_pred - y_true

        # moving the b inside the log allows for 0 values
        return float(np.mean(np.log((var_b / np.mean(var_b)) ** var_b) / np.mean(var_b)))



"""HGR Classifier.

Paper: "Fairness-Aware Learning for Continuous Attributes and Treatments, J. Mary, C. Calauzènes, N. El Karoui, ICML 2019."
Based on: https://github.com/criteo-research/continuous-fairness
http://proceedings.mlr.press/v97/mary19a/mary19a.pdf
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import List
from typing_extensions import Literal, TypedDict

from ranzen import implements

from ethicml.algorithms.inprocess.in_subprocess import InAlgorithmSubprocess
from ethicml.utility import HyperParamType

__all__ = ["HGR", "HgrArgs"]


class HgrArgs(TypedDict):
    """Args for the HGR implementation."""

    lr: float
    epochs: int
    mu: float
    batch_size: int
    model_type: Literal["deep_model", "linear_model"]


@dataclass
class HGR(InAlgorithmSubprocess):
    """HGR Method."""

    lr: float = 1e-3
    epochs: int = 50
    mu: float = 0.98
    batch_size: int = 128
    model_type: str = "deep_model"

    @implements(InAlgorithmSubprocess)
    def _get_path_to_script(self) -> List[str]:
        return ["-m", "ethicml.implementations.hgr_method"]

    @implements(InAlgorithmSubprocess)
    def _get_flags(self) -> HgrArgs:
        model_type: Literal["deep_model", "linear_model"] = (
            "deep_model" if self.model_type.lower() == "deep_model" else "linear_model"
        )
        return {
            "lr": self.lr,
            "epochs": self.epochs,
            "mu": self.mu,
            "batch_size": self.batch_size,
            "model_type": model_type,
        }

    @implements(InAlgorithmSubprocess)
    def get_hyperparameters(self) -> HyperParamType:
        return {
            "lr": self.lr,
            "epochs": self.epochs,
            "mu": self.mu,
            "batch_size": self.batch_size,
            "model_type": self.model_type,
        }

    @implements(InAlgorithmSubprocess)
    def get_name(self) -> str:
        return f"HGR {self.model_type}"



"""Fair Dummies Implementation."""
from __future__ import annotations

import json
import random
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from joblib import dump, load

from ethicml import DataTuple, SoftPrediction, SubgroupTuple
from ethicml.implementations.hgr_modules.hgr_impl import HgrClassLearner

if TYPE_CHECKING:
    from ethicml.algorithms.inprocess.hgr import HgrArgs
    from ethicml.algorithms.inprocess.in_subprocess import InAlgoArgs


def fit(train: DataTuple, args: HgrArgs, seed: int = 888) -> HgrClassLearner:
    """Fit a model.

    :param train:
    :param args:
    """
    try:
        import torch

        torch.manual_seed(seed)
    except ImportError as e:
        raise RuntimeError(
            "In order to use PyTorch, please install it following the instructions as https://pytorch.org/ . "
        ) from e

    random.seed(seed)
    np.random.seed(seed)

    model = HgrClassLearner(
        lr=args["lr"],
        epochs=args["epochs"],
        mu=args["mu"],
        cost_pred=torch.nn.CrossEntropyLoss(),
        in_shape=len(train.x.columns),
        out_shape=train.y.nunique(),
        batch_size=args["batch_size"],
        model_type=args["model_type"],
    )
    return model.fit(train, seed=seed)


def predict(model: HgrClassLearner, test: SubgroupTuple) -> np.ndarray:
    """Compute predictions on the given test data.

    :param exponentiated_gradient:
    :param test:
    """
    return model.predict(test.x)


def train_and_predict(
    train: DataTuple, test: SubgroupTuple, args: HgrArgs, seed: int
) -> np.ndarray:
    """Train a logistic regression model and compute predictions on the given test data.

    :param train:
    :param test:
    :param args:
    """
    model = fit(train, args, seed)
    return predict(model, test)


def main() -> None:
    """Run the Agarwal model as a standalone program."""
    in_algo_args: InAlgoArgs = json.loads(sys.argv[1])
    flags: HgrArgs = json.loads(sys.argv[2])

    if in_algo_args["mode"] == "run":
        train, test = DataTuple.from_npz(Path(in_algo_args["train"])), SubgroupTuple.from_npz(
            Path(in_algo_args["test"])
        )
        SoftPrediction(soft=train_and_predict(train, test, flags, in_algo_args["seed"])).to_npz(
            Path(in_algo_args["predictions"])
        )
    elif in_algo_args["mode"] == "fit":
        data = DataTuple.from_npz(Path(in_algo_args["train"]))
        model = fit(data, flags, in_algo_args["seed"])
        setattr(model, "ethicml_random_seed", in_algo_args["seed"])  # need to save the seed as well
        dump(model, Path(in_algo_args["model"]))
    elif in_algo_args["mode"] == "predict":
        test = SubgroupTuple.from_npz(Path(in_algo_args["test"]))
        model = load(Path(in_algo_args["model"]))
        SoftPrediction(soft=predict(model, test)).to_npz(Path(in_algo_args["predictions"]))
    else:
        raise RuntimeError(f"Unknown mode: {in_algo_args['mode']}")


if __name__ == "__main__":
    main()



"""Implementation of VFAE."""
from __future__ import annotations

import json
import sys
from pathlib import Path
from typing import TYPE_CHECKING, List, Tuple, Union

import pandas as pd
import torch
from joblib import dump, load
from torch import optim
from torch.optim import Adam
from torch.utils.data import DataLoader

from ethicml.data.lookup import get_dataset_obj_by_name
from ethicml.implementations.beutel import set_seed
from ethicml.utility import DataTuple, SubgroupTuple

from .pytorch_common import CustomDataset, TestDataset
from .utils import load_data_from_flags, save_transformations
from .vfae_modules import VFAENetwork, loss_function

if TYPE_CHECKING:
    from ethicml.algorithms.preprocess.pre_subprocess import PreAlgoArgs, T
    from ethicml.algorithms.preprocess.vfae import VfaeArgs


def fit(train: DataTuple, flags: VfaeArgs):
    """Train the model.

    :param train:
    :param flags:
    """
    dataset = get_dataset_obj_by_name(flags["dataset"])()

    # Set up the data
    train_data = CustomDataset(train)
    train_loader = DataLoader(train_data, batch_size=flags["batch_size"])

    # Build Network
    model = VFAENetwork(
        dataset,
        flags["supervised"],
        train_data.xdim,
        latent_dims=flags["latent_dims"],
        z1_enc_size=flags["z1_enc_size"],
        z2_enc_size=flags["z2_enc_size"],
        z1_dec_size=flags["z1_dec_size"],
    ).to("cpu")
    optimizer = optim.Adam(model.parameters(), lr=1e-3)

    # Run Network
    for epoch in range(int(flags["epochs"])):
        train_model(epoch, model, train_loader, optimizer, flags)
    return model


def transform(model: VFAENetwork, dataset: T, flags: VfaeArgs) -> T:
    """Transform the dataset.

    :param model:
    :param dataset:
    :param flags:
    """
    data: Union[CustomDataset, TestDataset]
    if isinstance(dataset, DataTuple):
        data = CustomDataset(dataset)
        loader = DataLoader(data, batch_size=flags["batch_size"], shuffle=False)
    elif isinstance(dataset, SubgroupTuple):
        data = TestDataset(dataset)
        loader = DataLoader(data, batch_size=flags["batch_size"], shuffle=False)

    post_train: List[List[float]] = []
    model.eval()
    with torch.no_grad():
        for sample in loader:
            if isinstance(dataset, DataTuple):
                _x, _s, _ = sample
            elif isinstance(dataset, SubgroupTuple):
                _x, _s = sample
            z1_mu, z1_logvar = model.encode_z1(_x, _s)
            # z1 = model.reparameterize(z1_mu, z1_logvar)
            post_train += z1_mu.data.tolist()

    return dataset.replace(x=pd.DataFrame(post_train)).rename(f"VFAE: {dataset.name}")


def train_and_transform(train: DataTuple, test: T, flags: VfaeArgs) -> Tuple[DataTuple, T]:
    """Train the model and transform the dataset.

    :param train:
    :param test:
    :param flags:
    :returns: Tuple of Encoded Train Dataset and Test Dataset.
    """
    model = fit(train, flags)

    # Transform output
    return transform(model, train, flags), transform(model, test, flags)


def train_model(
    epoch: int, model: VFAENetwork, train_loader: DataLoader, optimizer: Adam, flags: VfaeArgs
) -> None:
    """Train the model.

    :param epoch:
    :param model:
    :param train_loader:
    :param optimizer:
    :param flags:
    :returns: None
    """
    model.train()
    train_loss = 0.0
    for batch_idx, (data_x, data_s, data_y) in enumerate(train_loader):
        data_x = data_x.to("cpu")
        data_s = data_s.to("cpu")
        data_y = data_y.to("cpu")
        optimizer.zero_grad()
        z1_trip, z2_trip, z1_d_trip, x_dec, y_pred = model(data_x, data_s, data_y)

        data_trip = (data_x, data_s, data_y)

        loss_tuple = loss_function(flags, z1_trip, z2_trip, z1_d_trip, data_trip, x_dec, y_pred)
        prediction_loss, reconstruction_loss, kld_loss, mmd_loss = loss_tuple
        loss = kld_loss + reconstruction_loss + prediction_loss + mmd_loss
        loss.backward()
        train_loss += loss.item()
        optimizer.step()
        if batch_idx % 100 == 0:
            if flags["supervised"]:
                print(
                    f"train Epoch: {epoch} [{batch_idx * len(data_x)}/{len(train_loader.dataset)}"
                    f"({100. * batch_idx / len(train_loader):.0f}%)]\t"
                    f"Loss: {loss.item() / len(data_x):.6f}\t"
                    f"pred_loss: {prediction_loss.item():.6f}\t"
                    f"recon_loss: {reconstruction_loss.item():.6f}\t"
                    f"kld_loss: {kld_loss.item():.6f}\t"
                    f"mmd_loss: {flags['batch_size'] * mmd_loss.item():.6f}"
                )
            else:
                print(
                    f"train Epoch: {epoch} [{batch_idx * len(data_x)}/{len(train_loader.dataset)}"
                    f"({100. * batch_idx / len(train_loader):.0f}%)]\t"
                    f"Loss: {loss.item() / len(data_x):.6f}\t"
                    f"recon_loss: {reconstruction_loss.item():.6f}\t"
                    f"mmd_loss: {flags['batch_size'] * mmd_loss.item():.6f}"
                )

    print(f"====> Epoch: {epoch} Average loss: {train_loss / len(train_loader.dataset):.4f}")


def main() -> None:
    """Run model."""
    pre_algo_args: PreAlgoArgs = json.loads(sys.argv[1])
    flags: VfaeArgs = json.loads(sys.argv[2])
    if pre_algo_args["mode"] == "run":
        set_seed(pre_algo_args["seed"])
        train, test = load_data_from_flags(pre_algo_args)
        save_transformations(train_and_transform(train, test, flags), pre_algo_args)
    elif pre_algo_args["mode"] == "fit":
        set_seed(pre_algo_args["seed"])
        train = DataTuple.from_npz(Path(pre_algo_args["train"]))
        enc = fit(train, flags)
        transformed_train = transform(enc, train, flags)
        transformed_train.to_npz(Path(pre_algo_args["new_train"]))
        dump(enc, Path(pre_algo_args["model"]))
    elif pre_algo_args["mode"] == "transform":
        model = load(Path(pre_algo_args["model"]))
        transformed_test = transform(model, DataTuple.from_npz(Path(pre_algo_args["test"])), flags)
        transformed_test.to_npz(Path(pre_algo_args["new_test"]))


if __name__ == "__main__":
    main()



"""Manually specified (i.e. not learned) models."""
from dataclasses import dataclass

import numpy as np
import pandas as pd
from ranzen import implements

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmNoParams
from ethicml.utility import DataTuple, Prediction, TestTuple

__all__ = ["Corels"]


@dataclass
class Corels(InAlgorithmNoParams):
    """CORELS (Certifiably Optimal RulE ListS) algorithm for the COMPAS dataset.

    This algorithm uses if-statements to make predictions. It only works on COMPAS with s as sex.

    From this paper: https://arxiv.org/abs/1704.01701
    """

    @implements(InAlgorithmNoParams)
    def get_name(self) -> str:
        return "CORELS"

    @implements(InAlgorithmNoParams)
    def fit(self, train: DataTuple, seed: int = 888) -> "Corels":
        return self

    @implements(InAlgorithmNoParams)
    def predict(self, test: TestTuple) -> Prediction:
        if test.name is None or "Compas" not in test.name or test.s.name != "sex":
            raise RuntimeError("The Corels algorithm only works on the COMPAS dataset")
        age = test.x["age-num"].to_numpy()
        priors = test.x["priors-count"].to_numpy()
        sex = test.s.to_numpy()
        male = 1
        condition1 = (age >= 18) & (age <= 20) & (sex == male)
        condition2 = (age >= 21) & (age <= 23) & (priors >= 2) & (priors <= 3)
        condition3: np.ndarray = priors > 3
        pred = np.where(condition1 | condition2 | condition3, np.ones_like(age), np.zeros_like(age))
        return Prediction(hard=pd.Series(pred))

    @implements(InAlgorithmNoParams)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        return self.predict(test)



"""Class to describe features of the Adult dataset."""
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type, Union

from ..dataset import LoadableDataset
from ..util import LabelSpec, flatten_dict

__all__ = ["Nursery", "NurserySplits"]


class NurserySplits(Enum):
    """Available dataset splits for the Adult dataset."""

    FINANCE = "Finance"
    CUSTOM = "Custom"


@dataclass
class Nursery(LoadableDataset):
    """UCI Adult dataset.

    :param discrete_only: If True, continuous features are dropped. (Default: False)
    :param invert_s: If True, the (binary) ``s`` values are inverted. (Default: False)
    :param split: What to use as ``s``. (Default: "Sex")
    :param binarize_nationality: If True, nationality will be USA vs rest. (Default: False)
    :param binarize_race: If True, race will be white vs rest. (Default: False)
    """

    Splits: ClassVar[Type[NurserySplits]] = NurserySplits

    split: NurserySplits = Splits.FINANCE

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "parents": ["parents_usual", "parents_pretentious", "parents_great_pret"],
            "has_nurs": [
                "has_nurs_proper",
                "has_nurs_less_proper",
                "has_nurs_improper",
                "has_nurs_critical",
                "has_nurs_very_crit",
            ],
            "form": ["form_complete", "form_completed", "form_foster", "form_incomplete"],
            "housing": ["housing_convenient", "housing_less_conv", "housing_critical"],
            "health": ["health_not_recom", "health_priority", "health_recommended"],
            "social": ["social_non-prob", "social_slightly_prob", "social_problematic"],
            "class": [
                "class_not_recom",
                "class_recommend",
                "class_very_recom",
                "class_priority",
                "class_spec_prior",
            ],
            "finance": ["finance_conv", "finance_inconv"],
        }
        discrete_features = flatten_dict(disc_feature_groups)

        continuous_features = ["children"]

        sens_attr_spec: Union[str, LabelSpec]
        if self.split is NurserySplits.FINANCE:
            sens_attr_spec = "finance_conv"
            s_prefix = ["finance"]
            class_label_spec = "class_not_recom"
            class_label_prefix = ["class"]
        else:
            raise NotImplementedError

        name = f"Nursery {self.split.value}"

        super().__init__(
            name=name,
            num_samples=12960,
            features=discrete_features + continuous_features,
            cont_features=continuous_features,
            sens_attr_spec=sens_attr_spec,
            class_label_spec=class_label_spec,
            filename_or_path="nursery.csv.zip",
            s_prefix=s_prefix,
            class_label_prefix=class_label_prefix,
            discrete_only=self.discrete_only,
            discrete_feature_groups=disc_feature_groups,
        )



"""Class to describe features of the Lipton dataset."""
from dataclasses import dataclass

from ..dataset import LoadableDataset

__all__ = ["Lipton"]


@dataclass
class Lipton(LoadableDataset):
    """Synthetic dataset from the Lipton et al. 2018.

    Described in section 4.1 of `Does mitigating ML's impact disparity require treatment disparity?`

    .. code-block:: bibtex

        @article{lipton2018does,
            title={Does mitigating ML's impact disparity require treatment disparity?},
            author={Lipton, Zachary and McAuley, Julian and Chouldechova, Alexandra},
            journal={Advances in neural information processing systems},
            volume={31},
            pages={8125--8135},
            year={2018}
        }
    """

    def __post_init__(self) -> None:
        continuous_features = ["hair_length", "work_experience"]
        super().__init__(
            name="Lipton",
            num_samples=2_000,
            filename_or_path="lipton.csv",
            features=continuous_features,
            cont_features=continuous_features,
            sens_attr_spec="sens",
            class_label_spec="hired",
            discrete_only=False,
        )



"""Accuracy that is balanced with respect to the class labels."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric
from .metric import Metric

__all__ = ["BalancedAccuracy"]


@dataclass
class BalancedAccuracy(CfmMetric):
    """Accuracy that is balanced with respect to the class labels."""

    _name: ClassVar[str] = "Balanced Accuracy"

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        t_neg, f_pos, f_neg, t_pos = self.confusion_matrix(prediction=prediction, actual=actual)
        tpr = t_pos / (t_pos + f_neg)
        tnr = t_neg / (t_neg + f_pos)
        return 0.5 * (tpr + tnr)



"""Data structures that are used throughout the code."""
from __future__ import annotations

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    NewType,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)
from typing_extensions import Final, Literal, TypeAlias, final

import numpy as np
import pandas as pd
from numpy import typing as npt
from ranzen import enum_name_str

__all__ = [
    "ClassifierType",
    "DataTuple",
    "EvalTuple",
    "FairnessType",
    "HyperParamType",
    "HyperParamValue",
    "KernelType",
    "Prediction",
    "Results",
    "ResultsAggregator",
    "SoftPrediction",
    "SubgroupTuple",
    "TestTuple",
    "TrainTestPair",
    "TrainValPair",
    "aggregate_results",
    "concat",
    "filter_and_map_results",
    "filter_results",
    "make_results",
    "map_over_results_index",
]

AxisType: TypeAlias = Literal["columns", "index"]  # pylint: disable=invalid-name


class PandasIndex(Enum):
    """Enum for indexing the results."""

    DATASET = "dataset"
    SCALER = "scaler"
    TRANSFORM = "transform"
    MODEL = "model"


_S = TypeVar("_S", bound="SubsetMixin")


class SubsetMixin(ABC):
    """Mixin that provides methods for getting subsets."""

    # the mixin assumes the presence of these attributes
    data: pd.DataFrame
    s_column: str

    @abstractmethod
    def replace_data(self: _S, data: pd.DataFrame) -> _S:
        """Make a copy of the container but change the underlying data."""

    @property
    @final
    def s(self) -> pd.Series[int]:
        """Getter for property s."""
        return self.data[self.s_column]

    @final
    def get_n_samples(self: _S, num: int = 500) -> _S:
        """Get the first elements of the dataset.

        :param num: How many samples to take for subset. (Default: 500)
        :returns: Subset of training data.
        """
        return self.replace_data(data=self.data.iloc[:num])

    @final
    def get_s_subset(self: _S, s: int) -> _S:
        """Return a subset of the DataTuple where S=s."""
        return self.replace_data(data=self.data[self.s == s])

    @final
    def __len__(self) -> int:
        """Overwrite __len__ magic method."""
        return len(self.data)


@dataclass
class SubgroupTuple(SubsetMixin):
    """A tuple of dataframes for the features and the sensitive attribute."""

    __slots__ = ("data", "s_column", "name")
    data: pd.DataFrame
    s_column: str
    name: str | None

    def __post_init__(self) -> None:
        assert self.s_column in self.data.columns, f"column {self.s_column} not present"

    @classmethod
    def from_df(
        cls, *, x: pd.DataFrame, s: pd.Series[int], name: Optional[str] = None
    ) -> SubgroupTuple:
        """Make a SubgroupTuple."""
        s_column = s.name
        assert isinstance(s_column, str)
        assert len(x) == len(s), "data has to have the same length"
        return cls(data=pd.concat([x, s], axis="columns", sort=False), s_column=s_column, name=name)

    @property
    def x(self) -> pd.DataFrame:
        """Getter for property x."""
        return self.data.drop(self.s_column, inplace=False, axis="columns")

    def __iter__(self) -> Iterator[Union[pd.DataFrame, pd.Series]]:
        """Overwrite magic method __iter__."""
        return iter([self.x, self.s])

    def replace(
        self, *, x: Optional[pd.DataFrame] = None, s: Optional[pd.Series] = None
    ) -> SubgroupTuple:
        """Create a copy of the SubgroupTuple but change the given values."""
        return SubgroupTuple.from_df(
            x=x if x is not None else self.x, s=s if s is not None else self.s, name=self.name
        )

    def replace_data(self, data: pd.DataFrame) -> SubgroupTuple:
        """Make a copy of the DataTuple but change the underlying data."""
        assert self.s_column in data.columns, f"column {self.s_column} not present"
        return SubgroupTuple(data=data, s_column=self.s_column, name=self.name)

    def rename(self, name: str) -> SubgroupTuple:
        """Change only the name."""
        return SubgroupTuple(data=self.data, s_column=self.s_column, name=name)

    def to_npz(self, data_path: Path) -> None:
        """Save SubgroupTuple as an npz file.

        :param data_path: Path to save the npz file.
        """
        write_as_npz(
            data_path,
            dict(x=self.x, s=self.s),
            dict(name=np.array(self.name if self.name is not None else "")),
        )

    @classmethod
    def from_npz(cls, data_path: Path) -> SubgroupTuple:
        """Load test tuple from npz file.

        :param data_path: Path to load the npz file.
        """
        with data_path.open("rb") as data_file:
            data = np.load(data_file)
            name = data["name"].item()
            return cls.from_df(
                x=pd.DataFrame(data["x"], columns=data["x_names"]),
                s=pd.Series(data["s"], name=data["s_names"][0]),
                name=name or None,
            )


@dataclass
class DataTuple(SubsetMixin):
    """A tuple of dataframes for the features, the sensitive attribute and the class labels."""

    __slots__ = ("data", "s_column", "y_column", "name")
    data: pd.DataFrame
    s_column: str
    y_column: str
    name: str | None

    def __post_init__(self) -> None:
        assert self.s_column in self.data.columns, f"column {self.s_column} not present"
        assert self.y_column in self.data.columns, f"column {self.y_column} not present"

    @classmethod
    def from_df(
        cls, *, x: pd.DataFrame, s: pd.Series[int], y: pd.Series[int], name: Optional[str] = None
    ) -> DataTuple:
        """Make a DataTuple."""
        s_column = s.name
        y_column = y.name
        assert isinstance(s_column, str) and isinstance(y_column, str)
        assert len(x) == len(s) == len(y), "data has to have the same length"
        return cls(
            data=pd.concat([x, s, y], axis="columns", sort=False),
            s_column=s_column,
            y_column=y_column,
            name=name,
        )

    @property
    def x(self) -> pd.DataFrame:
        """Getter for property x."""
        return self.data.drop([self.s_column, self.y_column], inplace=False, axis="columns")

    @property
    def y(self) -> pd.Series[int]:
        """Getter for property y."""
        return self.data[self.y_column]

    def __iter__(self) -> Iterator[Union[pd.DataFrame, pd.Series]]:
        """Overwrite magic method __iter__."""
        return iter([self.x, self.s, self.y])

    def remove_y(self) -> SubgroupTuple:
        """Convert the DataTuple instance to a SubgroupTuple instance."""
        return SubgroupTuple(
            data=self.data.drop(self.y_column, inplace=False, axis="columns"),
            s_column=self.s_column,
            name=self.name,
        )

    def replace(
        self,
        *,
        x: Optional[pd.DataFrame] = None,
        s: Optional[pd.Series] = None,
        y: Optional[pd.Series] = None,
    ) -> DataTuple:
        """Create a copy of the DataTuple but change the given values."""
        return DataTuple.from_df(
            x=x if x is not None else self.x,
            s=s if s is not None else self.s,
            y=y if y is not None else self.y,
            name=self.name,
        )

    def rename(self, name: str) -> DataTuple:
        """Change only the name."""
        return DataTuple(data=self.data, s_column=self.s_column, y_column=self.y_column, name=name)

    def replace_data(self, data: pd.DataFrame) -> DataTuple:
        """Make a copy of the DataTuple but change the underlying data."""
        assert self.s_column in data.columns, f"column {self.s_column} not present"
        assert self.y_column in data.columns, f"column {self.y_column} not present"
        return DataTuple(data=data, s_column=self.s_column, y_column=self.y_column, name=self.name)

    def apply_to_joined_df(self, mapper: Callable[[pd.DataFrame], pd.DataFrame]) -> DataTuple:
        """Concatenate the dataframes in the DataTuple and then apply a function to it.

        :param mapper: A function that takes a dataframe and returns a dataframe.
        """
        return self.replace_data(data=mapper(self.data))

    def to_npz(self, data_path: Path) -> None:
        """Save DataTuple as an npz file.

        :param data_path: Path to the npz file.
        """
        write_as_npz(
            data_path,
            dict(x=self.x, s=self.s, y=self.y),
            dict(name=np.array(self.name if self.name is not None else "")),
        )

    @classmethod
    def from_npz(cls, data_path: Path) -> DataTuple:
        """Load data tuple from npz file.

        :param data_path: Path to the npz file.
        """
        with data_path.open("rb") as data_file:
            data = np.load(data_file)
            name = data["name"].item()
            return cls.from_df(
                x=pd.DataFrame(data["x"], columns=data["x_names"]),
                s=pd.Series(data["s"], name=data["s_names"][0]),
                y=pd.Series(data["y"], name=data["y_names"][0]),
                name=name or None,
            )


@dataclass
class LabelTuple(SubsetMixin):
    """A tuple of dataframes for the features, the sensitive attribute and the class labels."""

    __slots__ = ("data", "s_column", "y_column", "name")
    data: pd.DataFrame
    s_column: str
    y_column: str
    name: str | None

    def __post_init__(self) -> None:
        assert self.s_column in self.data.columns, f"column {self.s_column} not present"
        assert self.y_column in self.data.columns, f"column {self.y_column} not present"

    @classmethod
    def from_df(
        cls, *, s: pd.Series[int], y: pd.Series[int], name: Optional[str] = None
    ) -> LabelTuple:
        """Make a LabelTuple."""
        s_column = s.name
        y_column = y.name
        assert isinstance(s_column, str) and isinstance(y_column, str)
        assert len(s) == len(y), "data has to have the same length"
        return cls(
            data=pd.concat([s, y], axis="columns", sort=False),  # type: ignore[arg-type]
            s_column=s_column,
            y_column=y_column,
            name=name,
        )

    @classmethod
    def from_np(
        cls, *, s: npt.NDArray, y: npt.NDArray, s_name: str = "s", y_name: str = "y"
    ) -> LabelTuple:
        """Create a LabelTuple from numpy arrays."""
        s_pd = pd.Series(s, name=s_name)
        y_pd = pd.Series(y, name=y_name)
        assert len(s_pd) == len(y_pd), "data has to have the same length"
        return cls(
            data=pd.concat([s_pd, y_pd], axis="columns", sort=False),
            s_column=s_name,
            y_column=y_name,
            name=None,
        )

    @property
    def y(self) -> pd.Series[int]:
        """Getter for property y."""
        return self.data[self.y_column]

    def __iter__(self) -> Iterator[Union[pd.DataFrame, pd.Series]]:
        """Overwrite magic method __iter__."""
        return iter([self.s, self.y])

    def replace(
        self, *, s: Optional[pd.Series] = None, y: Optional[pd.Series] = None
    ) -> LabelTuple:
        """Create a copy of the LabelTuple but change the given values."""
        return LabelTuple.from_df(
            s=s if s is not None else self.s, y=y if y is not None else self.y, name=self.name
        )

    def rename(self, name: str) -> LabelTuple:
        """Change only the name."""
        return LabelTuple(data=self.data, s_column=self.s_column, y_column=self.y_column, name=name)

    def replace_data(self, data: pd.DataFrame) -> LabelTuple:
        """Make a copy of the LabelTuple but change the underlying data."""
        assert self.s_column in data.columns, f"column {self.s_column} not present"
        assert self.y_column in data.columns, f"column {self.y_column} not present"
        return LabelTuple(data=data, s_column=self.s_column, y_column=self.y_column, name=self.name)


TestTuple: TypeAlias = Union[SubgroupTuple, DataTuple]
EvalTuple: TypeAlias = Union[LabelTuple, DataTuple]
T = TypeVar("T", SubgroupTuple, DataTuple)


class Prediction:
    """Prediction of an algorithm."""

    def __init__(self, hard: pd.Series, info: Optional[HyperParamType] = None):
        """Make a prediction obj."""
        assert isinstance(hard, pd.Series), "please use pd.Series"
        self._hard = hard
        self._info = info if info is not None else {}

    @classmethod
    def from_np(cls, preds: npt.NDArray) -> Prediction:
        """Construct a prediction object from a numpy array."""
        return cls(hard=pd.Series(preds))

    def __len__(self) -> int:
        """Length of the predictions object."""
        return len(self._hard)

    def get_s_subset(self, s_data: pd.Series, s: int) -> Prediction:
        """Return a subset of the DataTuple where S=s.

        :param s_data: Dataframe with the s-values.
        :param s: S-value to get the subset for.
        """
        return Prediction(hard=self.hard[s_data == s])

    @property
    def hard(self) -> pd.Series:
        """Hard predictions (e.g. 0 and 1)."""
        return self._hard

    @property
    def info(self) -> HyperParamType:
        """Additional info about the prediction."""
        return self._info

    @staticmethod
    def from_npz(npz_path: Path) -> Prediction:
        """Load prediction from npz file.

        :param npz_path: Path to the npz file.
        """
        info = None
        if (npz_path.parent / "info.json").exists():
            with open(npz_path.parent / "info.json", encoding="utf-8") as json_file:
                info = json.load(json_file)
        with npz_path.open("rb") as npz_file:
            data = np.load(npz_file)
            if "soft" in data:
                return SoftPrediction(soft=np.squeeze(data["soft"]), info=info)
            return Prediction(hard=pd.Series(np.squeeze(data["hard"])), info=info)

    def to_npz(self, npz_path: Path) -> None:
        """Save prediction as npz file.

        :param npz_path: Path to the npz file.
        """
        if self.info:
            for v in self.info.values():
                assert isinstance(v, float), "Info must be Dict[str, float]"
            json_path = npz_path.parent / "info.json"
            with open(json_path, "w") as json_file:
                json.dump(self.info, json_file)

        np.savez(npz_path, hard=self.hard.to_numpy())


class SoftPrediction(Prediction):
    """Prediction of an algorithm that makes soft predictions."""

    def __init__(self, soft: np.ndarray, info: Optional[HyperParamType] = None):
        """Make a soft prediction object."""
        super().__init__(hard=pd.Series(soft.argmax(axis=1).astype(int), name="hard"), info=info)
        self._soft = soft

    @property
    def soft(self) -> np.ndarray:
        """Soft predictions (e.g. 0.2 and 0.8)."""
        return self._soft


def write_as_npz(
    data_path: Path,
    data: Dict[str, Union[pd.DataFrame, pd.Series]],
    extra: Optional[Dict[str, np.ndarray]] = None,
) -> None:
    """Write the given dataframes to an npz file.

    :param data_path: Path to the npz file.
    :param data: Dataframes to save.
    :param extra: Extra data to save. (Default: None)
    """
    extra = extra or {}
    as_numpy = {entry: values.to_numpy() for entry, values in data.items()}
    column_names: Dict[str, np.ndarray] = {
        f"{entry}_names": np.array(values.columns.tolist())
        if isinstance(values, pd.DataFrame)
        else np.array([values.name])
        for entry, values in data.items()
    }

    np.savez(data_path, **as_numpy, **column_names, **extra)


def concat(
    datatup_list: Sequence[T],
    ignore_index: bool = False,
) -> T:
    """Concatenate the data tuples in the given list.

    :param datatup_list: List of data tuples to concatenate.
    :param ignore_index: Ignore the index of the dataframes. (Default: False)
    """
    data: pd.DataFrame = pd.concat(
        [dt.data for dt in datatup_list], axis="index", sort=False, ignore_index=ignore_index
    )
    return datatup_list[0].replace_data(data)


@enum_name_str
class FairnessType(Enum):
    """Fairness type."""

    dp = auto()
    eq_opp = auto()
    eq_odds = auto()


@enum_name_str
class ClassifierType(Enum):
    """Classifier type."""

    lr = auto()
    svm = auto()
    gbt = auto()


class TrainTestPair(NamedTuple):
    """2-Tuple of train and test data."""

    train: DataTuple
    test: SubgroupTuple


class TrainValPair(NamedTuple):
    """2-Tuple of train and test data."""

    train: DataTuple
    test: DataTuple


Results = NewType("Results", pd.DataFrame)  # Container for results from `evaluate_models`

RESULTS_COLUMNS: Final = ["dataset", "scaler", "transform", "model", "split_id"]


def make_results(data_frame: Union[None, pd.DataFrame, Path] = None) -> Results:
    """Initialise Results object.

    You should always use this function instead of using the "constructor" directly, because this
    function checks whether the columns are correct.

    :param data_frame: A dataframe to use for initialization. (Default: None)
    """
    if isinstance(data_frame, Path):
        data_frame = pd.read_csv(data_frame)
    if data_frame is None:
        return Results(pd.DataFrame(columns=RESULTS_COLUMNS).set_index(RESULTS_COLUMNS))
    # ensure correct index
    if data_frame.index.names != RESULTS_COLUMNS:
        return Results(data_frame.set_index(RESULTS_COLUMNS))
    else:
        return Results(data_frame)


class ResultsAggregator:
    """Aggregate results."""

    def __init__(self, initial: Optional[pd.DataFrame] = None):
        """Init results aggregator obj."""
        self._results = make_results(initial)

    @property
    def results(self) -> Results:
        """Results object over which this class is aggregating."""
        return self._results

    def append_df(self, data_frame: pd.DataFrame, prepend: bool = False) -> None:
        """Append (or prepend) a DataFrame to this object.

        :param data_frame: DataFrame to append.
        :param prepend: Whether to prepend or append the dataframe. (Default: False)
        """
        if data_frame.index.names != RESULTS_COLUMNS:
            data_frame = data_frame.set_index(RESULTS_COLUMNS)  # set correct index
        order = [data_frame, self.results] if prepend else [self.results, data_frame]
        # set sort=False so that the order of the columns is preserved
        self._results = Results(pd.concat(order, sort=False, axis="index"))

    def append_from_csv(self, csv_file: Path, prepend: bool = False) -> bool:
        """Append results from a CSV file.

        :param csv_file: Path to the CSV file.
        :param prepend:  (Default: False)
        """
        if csv_file.is_file():  # if file exists
            self.append_df(pd.read_csv(csv_file), prepend=prepend)
            return True
        return False

    def save_as_csv(self, file_path: Path) -> None:
        """Save to csv.

        :param file_path: Path to the CSV file.
        """
        # `results` has the multi index based on [dataset, transform, ...] so we have to reset that
        self.results.reset_index(drop=False).to_csv(file_path, index=False)


def map_over_results_index(
    results: Results,
    mapper: Callable[[Tuple[str, str, str, str, str]], Tuple[str, str, str, str, str]],
) -> Results:
    """Change the values of the index with a transformation function."""
    results_mapped = results.copy()
    results_mapped.index = results_mapped.index.map(mapper)
    return make_results(results_mapped)


def filter_results(
    results: Results,
    values: Iterable,
    index: Union[str, PandasIndex] = "model",
) -> Results:
    """Filter the entries based on the given values.

    :param results: Results object to filter.
    :param values: Values to filter on.
    :param index: Index to filter on. (Default: "model")
    """
    if isinstance(index, str):
        index = PandasIndex(index)
    return Results(results.loc[results.index.get_level_values(index.value).isin(list(values))])


def filter_and_map_results(results: Results, mapping: Mapping[str, str]) -> Results:
    """Filter entries and change the index with a mapping.

    :param results: Results object to filter.
    :param mapping: Mapping from old index to new index.
    """
    return map_over_results_index(
        filter_results(results, mapping),
        lambda index: (index[0], index[1], index[2], mapping[index[3]], index[4]),
    )


def aggregate_results(
    results: Results, metrics: List[str], aggregator: Union[str, Tuple[str, ...]] = ("mean", "std")
) -> pd.DataFrame:
    """Aggregate results over the repeats.

    :param results: Results object containing the results to aggregate.
    :param metrics: Metrics used for aggregation.
    :param aggregator: Aggregator to use. The aggreators are the ones used in pandas.
        (Default: ("mean", "std"))
    """
    return results.groupby(["dataset", "scaler", "transform", "model"]).agg(aggregator)[metrics]  # type: ignore[arg-type]


@enum_name_str
class KernelType(Enum):
    """Values for SVM Kernel."""

    linear = auto()
    poly = auto()
    rbf = auto()
    sigmoid = auto()


HyperParamValue: TypeAlias = Union[bool, int, float, str, FairnessType, KernelType]
HyperParamType: TypeAlias = Dict[str, HyperParamValue]



"""For assessing FPR."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric
from .metric import Metric

__all__ = ["FPR"]


@dataclass
class FPR(CfmMetric):
    """False positive rate."""

    _name: ClassVar[str] = "FPR"

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        t_neg, f_pos, _, _ = self.confusion_matrix(prediction=prediction, actual=actual)
        return f_pos / (f_pos + t_neg)



"""Algorithms by Zafar et al. for Demographic Parity."""
import json
from abc import abstractmethod
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, ClassVar, Dict, List, NamedTuple, Optional, Union
from typing_extensions import Final

import pandas as pd
from ranzen import implements

from ethicml.preprocessing.adjust_labels import LabelBinarizer
from ethicml.utility import DataTuple, Prediction, TestTuple

from .in_algorithm import HyperParamType, InAlgorithm
from .installed_model import InstalledModel

__all__ = ["ZafarAccuracy", "ZafarBaseline", "ZafarEqOdds", "ZafarEqOpp", "ZafarFairness"]


SUB_DIR_IMPACT: Final = Path(".") / "disparate_impact" / "run-classifier"
SUB_DIR_MISTREAT: Final = Path(".") / "disparate_mistreatment" / "run_classifier"


class FitParams(NamedTuple):
    model_path: Path
    label_converter: LabelBinarizer


class _ZafarAlgorithmBase(InstalledModel):
    def __init__(self, name: str, sub_dir: Path):
        super().__init__(
            name=name,
            dir_name="zafar",
            url="https://github.com/predictive-analytics-lab/fair-classification.git",
            top_dir="fair-classification",
            use_poetry=True,
        )
        self._sub_dir = sub_dir
        self._fit_params: Optional[FitParams] = None

    @staticmethod
    def _create_file_in_zafar_format(
        data: Union[DataTuple, TestTuple], file_path: Path, label_converter: LabelBinarizer
    ) -> None:
        """Save a DataTuple as a JSON file, which is extremely inefficient but what Zafar wants.

        :param data: DataTuple to save.
        :param file_path: Path to save to.
        :param label_converter: Instance of a LabelBinarizer to convert labels to Zafar's format.
        """
        out: Dict[str, Any] = {'x': data.x.to_numpy().tolist(), "sensitive": {}}
        out["sensitive"][data.s.name] = data.s.to_numpy().tolist()
        if isinstance(data, DataTuple):
            data_converted = label_converter.adjust(data)
            out["class"] = (2 * data_converted.y.to_numpy() - 1).tolist()
        else:
            out["class"] = [-1 for _ in range(data.x.shape[0])]
        with file_path.open("w") as out_file:
            json.dump(out, out_file)

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            fit_params = self._fit(train, tmp_path, seed)
            return self._predict(test, tmp_path, fit_params)

    @implements(InAlgorithm)
    def fit(self, train: DataTuple, seed: int = 888) -> "_ZafarAlgorithmBase":
        with TemporaryDirectory() as tmpdir:
            self._fit_params = self._fit(
                train, tmp_path=Path(tmpdir), seed=seed, model_dir=self._code_path
            )
        return self

    @implements(InAlgorithm)
    def predict(self, test: TestTuple) -> Prediction:
        assert self._fit_params is not None, "call fit() first"
        with TemporaryDirectory() as tmpdir:
            return self._predict(test, tmp_path=Path(tmpdir), fit_params=self._fit_params)

    def _fit(
        self, train: DataTuple, tmp_path: Path, seed: int, model_dir: Optional[Path] = None
    ) -> FitParams:
        model_path = (model_dir.resolve() if model_dir is not None else tmp_path) / "model.npy"
        label_converter = LabelBinarizer()
        train_path = tmp_path / "train.json"
        self._create_file_in_zafar_format(train, train_path, label_converter)

        cmd = self._get_fit_cmd(str(train_path), str(model_path))
        working_dir = self._code_path.resolve() / self._sub_dir
        self.call_script(cmd, cwd=working_dir)

        return FitParams(model_path, label_converter)

    def _predict(self, test: TestTuple, tmp_path: Path, fit_params: FitParams) -> Prediction:
        test_path = tmp_path / "test.json"
        self._create_file_in_zafar_format(test, test_path, fit_params.label_converter)
        predictions_path = tmp_path / "predictions.json"
        cmd = self._get_predict_cmd(
            str(test_path), str(fit_params.model_path), str(predictions_path)
        )
        working_dir = self._code_path.resolve() / self._sub_dir
        self.call_script(cmd, cwd=working_dir)
        predictions = predictions_path.open().read()
        predictions = json.loads(predictions)

        predictions_correct = pd.Series([0 if x == -1 else 1 for x in predictions])
        return Prediction(hard=fit_params.label_converter.post_only_labels(predictions_correct))

    @abstractmethod
    def _get_fit_cmd(self, train_name: str, model_path: str) -> List[str]:
        pass

    def _get_predict_cmd(self, test_name: str, model_path: str, output_file: str) -> List[str]:
        return ["predict.py", test_name, model_path, output_file]


class ZafarBaseline(_ZafarAlgorithmBase):
    """Zafar without fairness."""

    is_fairness_algo: ClassVar[bool] = False

    def __init__(self) -> None:
        super().__init__(name="ZafarBaseline", sub_dir=SUB_DIR_IMPACT)

    @implements(InAlgorithm)
    def get_hyperparameters(self) -> HyperParamType:
        return {}

    @implements(_ZafarAlgorithmBase)
    def _get_fit_cmd(self, train_name: str, model_path: str) -> List[str]:
        return ["fit.py", train_name, model_path, "baseline", "0"]


class ZafarAccuracy(_ZafarAlgorithmBase):
    """Zafar with fairness."""

    def __init__(self, *, gamma: float = 0.5):
        super().__init__(name=f"ZafarAccuracy, γ={gamma}", sub_dir=SUB_DIR_IMPACT)
        self.gamma = gamma

    @implements(InAlgorithm)
    def get_hyperparameters(self) -> HyperParamType:
        return {"gamma": self.gamma}

    @implements(_ZafarAlgorithmBase)
    def _get_fit_cmd(self, train_name: str, model_path: str) -> List[str]:
        return ["fit.py", train_name, model_path, "gamma", str(self.gamma)]


class ZafarFairness(_ZafarAlgorithmBase):
    """Zafar with fairness."""

    def __init__(self, *, C: float = 0.001):
        super().__init__(name=f"ZafarFairness, C={C}", sub_dir=SUB_DIR_IMPACT)
        self._c = C

    @implements(InAlgorithm)
    def get_hyperparameters(self) -> HyperParamType:
        return {"C": self._c}

    @implements(_ZafarAlgorithmBase)
    def _get_fit_cmd(self, train_name: str, model_path: str) -> List[str]:
        return ["fit.py", train_name, model_path, "c", str(self._c)]


class ZafarEqOpp(_ZafarAlgorithmBase):
    """Zafar for Equality of Opportunity."""

    _mode: ClassVar[str] = "fnr"  # class level constant
    _base_name: ClassVar[str] = "ZafarEqOpp"

    def __init__(self, *, tau: float = 5.0, mu: float = 1.2, eps: float = 0.0001):
        name = f"{self._base_name}, τ={tau}, μ={mu} ε={eps}"
        super().__init__(name=name, sub_dir=SUB_DIR_MISTREAT)
        self._tau = tau
        self._mu = mu
        self._eps = eps

    @implements(InAlgorithm)
    def get_hyperparameters(self) -> HyperParamType:
        return {"tau": self._tau, "mu": self._mu, "eps": self._eps}

    @implements(_ZafarAlgorithmBase)
    def _get_fit_cmd(self, train_name: str, model_path: str) -> List[str]:
        return [
            "fit.py",
            train_name,
            model_path,
            self._mode,
            str(self._tau),
            str(self._mu),
            str(self._eps),
        ]


class ZafarEqOdds(ZafarEqOpp):
    """Zafar for Equalised Odds."""

    _mode: ClassVar[str] = "fprfnr"
    _base_name: ClassVar[str] = "ZafarEqOdds"



"""Base Class for decision / activation functions."""

from abc import ABC, abstractmethod

import numpy

__all__ = ["Activation"]


class Activation(ABC):
    """Base class for decision functions."""

    @abstractmethod
    def apply(self, soft_output: numpy.ndarray) -> numpy.ndarray:
        """Apply the decision function to a soft prediction.

        :param soft_output: soft prediction (i.e. a probability or logits)
        :returns: decision
        """

    @abstractmethod
    def get_name(self) -> str:
        """Name of activation function."""



"""Class to describe features of the Synthetic dataset."""
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type

import teext as tx

from ..dataset import LoadableDataset

__all__ = ["SyntheticScenarios", "SyntheticTargets", "Synthetic"]


class SyntheticScenarios(Enum):
    """Scenarios for the synthetic dataset."""

    S1 = 1
    S2 = 2
    S3 = 3
    S4 = 4


class SyntheticTargets(Enum):
    """Targets for the synthetic dataset."""

    Y1 = 1
    Y2 = 2
    Y3 = 3


@dataclass
class Synthetic(LoadableDataset):
    r"""Dataset with synthetic data.

    ⊥ = is independent of
    ~ = is an ancestor of in the causal model used to generate the data

    Scenario 1 = X⊥S & Y⊥S.
        - This models completely fair data.
    Scenario 2 = X_2⊥S & Y_2⊥S; X_1~S, Y_1~S & Y_3~S
        - This models data where the inputs are biased. This is propogated through to the target.
    Scenario 3 = X⊥S, Y_1⊥S, Y_2⊥S; Y_3~S
        - This models data where the target is biased.
    Scenario 4 = X_2⊥S, Y_2⊥S; X_1~S, Y_1~S, Y_3~S
        - This models data where both the input and target are directly biased.
    """

    Scenarios: ClassVar[Type[SyntheticScenarios]] = SyntheticScenarios
    Targets: ClassVar[Type[SyntheticTargets]] = SyntheticTargets

    scenario: SyntheticScenarios = SyntheticScenarios.S1
    target: SyntheticTargets = SyntheticTargets.Y3
    fair: bool = False
    num_samples: int = 1_000

    def __post_init__(self) -> None:
        scenario: int = self.scenario.value
        target: int = self.target.value
        assert self.num_samples <= 100_000
        num_samples = tx.assert_positive_int(self.num_samples)

        super().__init__(
            name=f"Synthetic - Scenario {scenario}, target {target}"
            + (" fair" if self.fair else ""),
            num_samples=num_samples,
            filename_or_path=f"synthetic_scenario_{scenario}.csv",
            features=["x1f", "x2f", "n1", "n2"] if self.fair else ["x1", "x2", "n1", "n2"],
            cont_features=["x1f", "x2f", "n1", "n2"] if self.fair else ["x1", "x2", "n1", "n2"],
            sens_attr_spec="s",
            class_label_spec=f"y{target}" + ("f" if self.fair else ""),
            discrete_only=self.discrete_only,
        )



"""Simply returns the majority label from the train set."""
from dataclasses import dataclass
from typing import ClassVar

import pandas as pd
from ranzen import implements

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmNoParams
from ethicml.utility import DataTuple, Prediction, TestTuple

__all__ = ["Majority"]


@dataclass
class Majority(InAlgorithmNoParams):
    """Simply returns the majority label from the train set."""

    is_fairness_algo: ClassVar[bool] = False

    @implements(InAlgorithmNoParams)
    def get_name(self) -> str:
        return "Majority"

    @implements(InAlgorithmNoParams)
    def fit(self, train: DataTuple, seed: int = 888) -> "Majority":
        self.maj = train.y.mode(dropna=True).to_numpy()
        return self

    @implements(InAlgorithmNoParams)
    def predict(self, test: TestTuple) -> Prediction:
        return Prediction(hard=pd.Series(self.maj.repeat(len(test.x))))

    @implements(InAlgorithmNoParams)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        maj = train.y.mode(dropna=True).to_numpy()
        return Prediction(hard=pd.Series(maj.repeat(len(test.x))))



"""Transparently show how the UCI Adult dataset was generated from the raw download."""

import numpy as np
import pandas as pd


def run_generate_adult() -> None:
    """Generate the UCI Adult dataset from scratch."""
    # Load the data
    train = pd.read_csv("raw/adult.data", header=None)
    test = pd.read_csv("raw/adult.test", skiprows=[0], header=None)

    # Give data column names
    columns = [
        "age",
        "workclass",
        "fnlwgt",
        "education",
        "education-num",
        "marital-status",
        "occupation",
        "relationship",
        "race",
        "sex",
        "capital-gain",
        "capital-loss",
        "hours-per-week",
        "native-country",
        "salary",
    ]

    train.columns = pd.Index(columns)
    test.columns = pd.Index(columns)

    # Concat the data
    all_data = pd.concat([train, test], axis=0)

    for col in all_data.columns:
        if all_data[col].dtype == np.object:  # type: ignore[attr-defined]
            all_data[col] = all_data[col].str.strip()

    # Replace full stop in the label of the test set
    all_data = all_data.replace("<=50K.", np.str_("<=50K"))
    all_data = all_data.replace(">50K.", np.str_(">50K"))

    # Drop NaNs
    all_data = all_data.replace(r"^\s*\?+\s*$", np.nan, regex=True).dropna()

    # OHE
    all_data = pd.get_dummies(all_data)

    # Shuffle the data
    all_data = all_data.sample(frac=1.0, random_state=888).reset_index(drop=True)

    # Save the CSV
    all_data.to_csv("./adult.csv", index=False)


if __name__ == "__main__":
    run_generate_adult()



"""Utility Functins for HGR method."""
import torch


def compute_acc(yhat: torch.Tensor, y: torch.Tensor) -> float:
    """Accuracy."""
    _, predicted = torch.max(yhat, 1)
    total = y.size(0)
    correct = (predicted == y).sum().item()
    return correct / total



"""Set of scripts for splitting the train and test datasets based on conditions."""

from typing import Tuple

import pandas as pd

from ethicml.utility import DataTuple

__all__ = ["dataset_from_cond", "domain_split", "query_dt"]


def dataset_from_cond(dataset: pd.DataFrame, cond: str) -> pd.DataFrame:
    """Return the dataframe that meets some condition."""
    return dataset.query(cond).reset_index(drop=True)


def query_dt(datatup: DataTuple, query_str: str) -> DataTuple:
    """Query a datatuple."""
    assert isinstance(query_str, str)
    assert isinstance(datatup, DataTuple)

    def _query_func(joined_data_frame: pd.DataFrame) -> pd.DataFrame:
        return dataset_from_cond(joined_data_frame, cond=query_str)

    return datatup.apply_to_joined_df(_query_func)


def domain_split(
    datatup: DataTuple, tr_cond: str, te_cond: str, seed: int = 888
) -> Tuple[DataTuple, DataTuple]:
    """Split a datatuple based on a condition.

    :param datatup: DataTuple
    :param tr_cond: condition for the training set
    :param te_cond: condition for the test set
    :param seed:  (Default: 888)
    :returns: Tuple of DataTuple split into train and test. The test is all those that meet
        the test condition plus the same percentage again of the train set.
    """
    dataset = datatup.x

    train_dataset = dataset_from_cond(dataset, tr_cond)
    test_dataset = dataset_from_cond(dataset, te_cond)

    assert train_dataset.shape[0] + test_dataset.shape[0] == dataset.shape[0]

    test_pct = test_dataset.shape[0] / dataset.shape[0]
    train_pct = 1 - test_pct

    train_train_pcnt = (1 - (test_pct * 2)) / train_pct

    train_train = train_dataset.sample(frac=train_train_pcnt, random_state=seed)
    test_train = train_dataset.drop(train_train.index, axis="index")

    test = pd.concat([test_train, test_dataset], axis="index")

    train_x = datatup.x.iloc[train_train.index].reset_index(drop=True)
    train_s = datatup.s.iloc[train_train.index].reset_index(drop=True)
    train_y = datatup.y.iloc[train_train.index].reset_index(drop=True)

    train_datatup = DataTuple.from_df(x=train_x, s=train_s, y=train_y, name=datatup.name)

    test_x = datatup.x.iloc[test.index].reset_index(drop=True)
    test_s = datatup.s.iloc[test.index].reset_index(drop=True)
    test_y = datatup.y.iloc[test.index].reset_index(drop=True)

    test_datatup = DataTuple.from_df(x=test_x, s=test_s, y=test_y, name=datatup.name)

    return train_datatup, test_datatup



"""Simple upsampler that makes subgroups the same size as the majority group."""
import itertools
from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, List, Optional, Tuple

import pandas as pd
from ranzen import enum_name_str, implements

from ethicml.utility import DataTuple, SoftPrediction

from ..inprocess.logistic_regression import LR
from .pre_algorithm import PreAlgorithm, T

__all__ = ["Upsampler", "UpsampleStrategy"]


@enum_name_str
class UpsampleStrategy(Enum):
    """Strategy for upsampling."""

    uniform = auto()
    preferential = auto()
    naive = auto()


@dataclass
class Upsampler(PreAlgorithm):
    """Upsampler algorithm.

    Given a datatuple, create a larger datatuple such that the subgroups have a balanced number
    of samples.
    """

    strategy: UpsampleStrategy = UpsampleStrategy.uniform

    def __post_init__(self) -> None:
        self._out_size: Optional[int] = None

    @implements(PreAlgorithm)
    def get_name(self) -> str:
        return f"Upsample {self.strategy}"

    @implements(PreAlgorithm)
    def get_out_size(self) -> int:
        assert self._out_size is not None
        return self._out_size

    @implements(PreAlgorithm)
    def fit(self, train: DataTuple, seed: int = 888) -> Tuple["Upsampler", DataTuple]:
        self._out_size = train.x.shape[1]
        new_train, _ = upsample(train, train, self.strategy, seed, name=self.name)
        return self, new_train

    @implements(PreAlgorithm)
    def transform(self, data: T) -> T:
        return data.rename(f"{self.name}: {data.name}")

    @implements(PreAlgorithm)
    def run(self, train: DataTuple, test: T, seed: int = 888) -> Tuple[DataTuple, T]:
        self._out_size = train.x.shape[1]
        return upsample(train, test, self.strategy, seed, name=self.name)


def concat_datatuples(first_dt: DataTuple, second_dt: DataTuple) -> DataTuple:
    """Given 2 datatuples, concatenate them and shuffle."""
    assert (first_dt.x.columns == second_dt.x.columns).all()  # type: ignore[attr-defined]
    assert first_dt.s_column == second_dt.s_column
    assert first_dt.y_column == second_dt.y_column

    s_column = first_dt.s_column
    y_column = first_dt.y_column

    a_combined: pd.DataFrame = first_dt.data
    b_combined: pd.DataFrame = second_dt.data

    combined = pd.concat([a_combined, b_combined], axis="index")
    combined: pd.DataFrame = combined.sample(frac=1.0, random_state=1).reset_index(drop=True)  # type: ignore[assignment]

    return DataTuple(data=combined, s_column=s_column, y_column=y_column, name=first_dt.name)


def upsample(
    dataset: DataTuple,
    test: T,
    strategy: UpsampleStrategy,
    seed: int,
    name: str,
) -> Tuple[DataTuple, T]:
    """Upsample a datatuple.

    :param dataset: Dataset that is used to determine the imbalance.
    :param test: Test set that is upsampled along with ``dataset``.
    :param strategy: Strategy to use for upsampling.
    :param seed: Seed for the upsampling.
    :param name: Name of the upsampling strategy.
    """
    s_vals: List[int] = list(map(int, dataset.s.unique()))
    y_vals: List[int] = list(map(int, dataset.y.unique()))

    groups = itertools.product(s_vals, y_vals)

    data: Dict[Tuple[int, int], DataTuple] = {}
    for s, y in groups:
        s_y_mask = (dataset.s == s) & (dataset.y == y)
        data[(s, y)] = dataset.replace_data(dataset.data.loc[s_y_mask].reset_index(drop=True))

    percentages: Dict[Tuple[int, int], float] = {}

    vals: List[int] = []
    for key, val in data.items():
        vals.append(val.x.shape[0])

    for key, val in data.items():
        if strategy is UpsampleStrategy.naive:
            percentages[key] = max(vals) / val.x.shape[0]
        else:
            s_val: int = key[0]
            y_val: int = key[1]

            y_eq_y = dataset.y.loc[dataset.y == y_val].count()
            s_eq_s = dataset.s.loc[dataset.s == s_val].count()

            num_samples = dataset.y.count()
            num_batch = val.y.count()

            percentages[key] = round((y_eq_y * s_eq_s / (num_batch * num_samples)), 8)

    upsampled: Dict[Tuple[int, int], DataTuple] = {}
    all_data: pd.DataFrame
    for key, val in data.items():
        all_data = val.data.sample(  # type: ignore[assignment]
            frac=percentages[key], random_state=seed, replace=True
        ).reset_index(drop=True)
        upsampled[key] = val.replace_data(all_data)

    upsampled_datatuple: Optional[DataTuple] = None
    for key, val in upsampled.items():
        if upsampled_datatuple is None:
            upsampled_datatuple = val
        else:
            upsampled_datatuple = concat_datatuples(upsampled_datatuple, val)
            upsampled_datatuple = upsampled_datatuple.rename(f"{name}: {dataset.name}")

    if strategy is UpsampleStrategy.preferential:
        ranker = LR()
        rank: SoftPrediction = ranker.run(dataset, dataset)

        selected: List[pd.DataFrame] = []

        all_data = dataset.data
        all_data = pd.concat(
            [all_data, pd.DataFrame(rank.soft[:, 1], columns=["preds"])], axis="columns"
        )

        for key in data:

            s_val = key[0]
            y_val = key[1]
            s_y_mask = (dataset.s == s_val) & (dataset.y == y_val)

            ascending = s_val <= 0

            if percentages[key] > 1.0:
                selected.append(all_data.loc[s_y_mask])
                percentages[key] -= 1.0

            weight = all_data.loc[s_y_mask][dataset.y_column].count()
            selected.append(
                all_data.loc[s_y_mask]
                .sort_values(by=["preds"], ascending=ascending)
                .iloc[: int(percentages[key] * weight)]
            )

        upsampled_dataframes: pd.DataFrame
        for i, df in enumerate(selected):
            if i == 0:
                upsampled_dataframes = df.drop(["preds"], axis="columns")
            else:
                upsampled_dataframes = pd.concat(
                    [upsampled_dataframes, df.drop(["preds"], axis="columns")], axis="index"
                ).reset_index(drop=True)
        upsampled_datatuple = DataTuple(
            data=upsampled_dataframes,
            s_column=dataset.s_column,
            y_column=dataset.y_column,
            name=f"{name}: {dataset.name}",
        )

    assert upsampled_datatuple is not None
    return upsampled_datatuple, test.rename(name=f"{name}: {test.name}")



"""Evaluator for a metric per sensitive attribute class."""

from typing import Dict

import pandas as pd

from ethicml.utility import DataTuple, EvalTuple, Prediction, SoftPrediction

from .metric import Metric

__all__ = [
    "metric_per_sensitive_attribute",
    "diff_per_sensitive_attribute",
    "ratio_per_sensitive_attribute",
    "MetricNotApplicable",
]


class MetricNotApplicable(Exception):
    """Metric Not Applicable per sensitive attribute, apply to whole dataset instead."""


def metric_per_sensitive_attribute(
    prediction: Prediction, actual: EvalTuple, metric: Metric, use_sens_name: bool = True
) -> Dict[str, float]:
    """Compute a metric repeatedly on subsets of the data that share a senstitive attribute.

    :param prediction:
    :param actual:
    :param metric:
    :param use_sens_name:  (Default: True)
    """
    if not metric.apply_per_sensitive:
        raise MetricNotApplicable(
            f"Metric {metric.name} is not applicable per sensitive "
            f"attribute, apply to whole dataset instead"
        )

    assert actual.s.shape[0] == actual.y.shape[0]
    assert prediction.hard.shape[0] == actual.y.shape[0]

    per_sensitive_attr: Dict[str, float] = {}

    s_column: str = actual.s_column

    for unique_s in actual.s.unique():
        mask: pd.Series = actual.s == unique_s
        subset = actual.get_s_subset(unique_s)
        pred_y: Prediction
        if isinstance(prediction, SoftPrediction):
            pred_y = SoftPrediction(soft=prediction.soft[mask], info=prediction.info)
        else:
            pred_y = Prediction(
                hard=prediction.hard.loc[mask].reset_index(drop=True), info=prediction.info
            )
        key = (s_column if use_sens_name else "S") + "_" + str(unique_s)
        per_sensitive_attr[key] = metric.score(pred_y, subset)

    return per_sensitive_attr


def diff_per_sensitive_attribute(per_sens_res: Dict[str, float]) -> Dict[str, float]:
    """Compute the difference in the metrics per sensitive attribute.

    :param per_sens_res: dictionary of the results
    :returns: dictionary of differences
    """
    sens_values = sorted(per_sens_res.keys())
    diff_per_sens = {}

    for i, _ in enumerate(sens_values):
        i_value: float = per_sens_res[sens_values[i]]
        for j in range(i + 1, len(sens_values)):
            key: str = f"{sens_values[i]}-{sens_values[j]}"
            j_value: float = per_sens_res[sens_values[j]]
            diff_per_sens[key] = abs(i_value - j_value)

    return diff_per_sens


def ratio_per_sensitive_attribute(per_sens_res: Dict[str, float]) -> Dict[str, float]:
    """Compute the ratios in the metrics per sensitive attribute.

    :param per_sens_res: dictionary of the results
    :returns: dictionary of ratios
    """
    sens_values = sorted(per_sens_res.keys())
    ratio_per_sens = {}

    for i, _ in enumerate(sens_values):
        i_value: float = per_sens_res[sens_values[i]]
        for j in range(i + 1, len(sens_values)):
            key: str = f"{sens_values[i]}/{sens_values[j]}"
            j_value: float = per_sens_res[sens_values[j]]

            min_val = min(i_value, j_value)
            max_val = max(i_value, j_value)

            ratio_per_sens[key] = min_val / max_val if max_val != 0 else float("nan")

    return ratio_per_sens



"""Class to describe features of the Toy dataset."""

from dataclasses import dataclass

from ..dataset import LoadableDataset
from ..util import flatten_dict

__all__ = ["Toy"]


@dataclass
class Toy(LoadableDataset):
    """Dataset with toy data for testing."""

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "disc_1": ["disc_1_a", "disc_1_b", "disc_1_c", "disc_1_d", "disc_1_e"],
            "disc_2": ["disc_2_x", "disc_2_y", "disc_2_z"],
        }
        discrete_features = flatten_dict(disc_feature_groups)
        continuous_features = ["a1", "a2"]
        super().__init__(
            name="Toy",
            num_samples=400,
            filename_or_path="toy.csv",
            features=continuous_features + discrete_features,
            cont_features=continuous_features,
            sens_attr_spec="sensitive-attr",
            class_label_spec="decision",
            discrete_only=False,
            discrete_feature_groups=disc_feature_groups,
        )



"""Cross Validation for any in process (at the moment) Algorithm."""
from collections import defaultdict
from itertools import product
from statistics import mean
from typing import Any, Dict, List, Mapping, NamedTuple, Optional, Sequence, Tuple, Type

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm
from ethicml.evaluators.parallelism import run_in_parallel
from ethicml.metrics.accuracy import Accuracy
from ethicml.metrics.cv import AbsCV
from ethicml.metrics.metric import Metric
from ethicml.preprocessing.splits import fold_data
from ethicml.utility import DataTuple, Prediction, TrainValPair

__all__ = ["CrossValidator", "CVResults"]


class ResultTuple(NamedTuple):
    """Result of one experiment."""

    params: Dict[str, Any]  # the parameter setting used for this run
    fold_id: int  # the ID of the fold
    scores: Dict[str, float]  # the achieved scores


class CVResults:
    """Stores the results of a cross validation experiment.

    This object isn't meant to be iterated over directly.
    Instead, use the `raw_storage` property to access the results across all folds.
    Or, use the `mean_storage` property to access the average results for each parameter setting.


    .. code-block:: python

        import ethicml as em

        train, test = em.train_test_split(em.Compas().load())
        hyperparams = {"C": [1, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6]}

        cv = em.CrossValidator(em.LR, hyperparams, folds=3)
        primary = em.Accuracy()
        fair_measure = em.AbsCV()
        cv_results = cv.run(train, measures=[primary, fair_measure])
        best_result = cv_results.get_best_in_top_k(primary, fair_measure, top_k=3)

        print(f"Best C: {best_result.params['C']}")
        print(f"Best Accuracy: {best_result.scores['Accuracy']}")
        print(f"Best CV Score: {best_result.scores['CV absolute']}")
        print(cv_results.mean_storage)
        print(cv_results.raw_storage)

    """

    def __init__(self, results: List[ResultTuple], model: Type[InAlgorithm]):
        self.raw_storage = results
        self.model = model
        self.mean_storage = self._organize_and_compute_means()

    def _organize_and_compute_means(self) -> Dict[str, ResultTuple]:
        """Compute means over folds and generate unique string for each hyperparameter setting."""
        # first, group the entries that have the same hyperparameters
        max_fold_id = 0
        grouped: Dict[str, List[ResultTuple]] = defaultdict(list)
        for result in self.raw_storage:
            # we convert the hyperparameter dictionaries to strings in order to compare them
            hyp_string = ", ".join(f"{key!s}={val!r}" for (key, val) in result.params.items())
            grouped[hyp_string].append(result)
            if result.fold_id > max_fold_id:
                max_fold_id = result.fold_id

        # compute the mean value of each measure within each group
        mean_vals: Dict[str, ResultTuple] = {}
        for hyp_string, results in grouped.items():
            # re-order the data: we want a dictionary that has all the scores as a list
            # first we create empty list according to the dictionary keys in the first result
            scores_dict: Dict[str, List[float]] = {
                score_name: [] for score_name in results[0].scores
            }
            # then we iterate over results and score names to fill the lists of score values
            for result in results:
                for score_name, score in result.scores.items():
                    scores_dict[score_name].append(score)

            # the re-ordering is complete. now we can compute the mean
            mean_dict = {score_name: mean(scores) for score_name, scores in scores_dict.items()}
            # we save everything as a ResultTuple. as we don't have a fold id, we use -1
            mean_vals[hyp_string] = ResultTuple(results[0].params, -1, mean_dict)

        assert len(mean_vals) * (max_fold_id + 1) == len(self.raw_storage)
        assert len(list(mean_vals.values())[-1].scores) == len(self.raw_storage[-1].scores)
        return mean_vals

    def get_best_result(self, measure: Metric) -> ResultTuple:
        """Get the hyperparameter combination for the best performance of a measure.

        :param measure:
        """
        mean_vals = self.mean_storage

        def _get_score(item: Tuple[str, ResultTuple]) -> float:
            """Take an entry from `mean_storage` and return the desired score `measure`.

            :param item:
            """
            _, result = item
            return result.scores[measure.name]

        # find the best entry in `mean_storage` according to `measure`
        best_hyp_string, _ = max(mean_vals.items(), key=_get_score)

        return mean_vals[best_hyp_string]

    def best_hyper_params(self, measure: Metric) -> Dict[str, Any]:
        """Get hyper-parameters that return the 'best' result for the metric of interest.

        :param measure:
        """
        return self.get_best_result(measure).params

    def best(self, measure: Metric) -> InAlgorithm:
        """Return a model initialised with the hyper-parameters that perform optimally on average across folds for a given metric.

        :param measure:
        """
        return self.model(**self.best_hyper_params(measure))

    def get_best_in_top_k(self, primary: Metric, secondary: Metric, top_k: int) -> ResultTuple:
        """Get best result in top-K entries.

        First sort the results according to the primary metric, then take the best according to the
        secondary metric from the top K.

        :param primary: Metric to first sort by.
        :param secondary: Metric to sort the top-K models by for a second time, the top will be selected.
        :param top_k: Number of entries to consider.
        """
        mean_vals = self.mean_storage

        def _get_primary_score(item: Tuple[str, ResultTuple]) -> float:
            return item[1].scores[primary.name]

        sorted_by_primary = sorted(mean_vals.items(), key=_get_primary_score)
        top_k_candidates = sorted_by_primary[:top_k]

        def _get_secondary_score(item: Tuple[str, ResultTuple]) -> float:
            return item[1].scores[secondary.name]

        best_hyp_string, _ = max(top_k_candidates, key=_get_secondary_score)
        return mean_vals[best_hyp_string]


class _ResultsAccumulator:
    def __init__(self, measures: Optional[List[Metric]] = None):
        self._measures = [Accuracy(), AbsCV()] if measures is None else measures
        self.results: List[ResultTuple] = []

    def __call__(
        self, parameter_setting: Dict[str, Any], preds: Prediction, test: DataTuple, fold_id: int
    ) -> Dict[str, float]:
        """Compute the scores for the given predictions and append to the list of results."""
        # compute all measures
        # TODO: this should also compute diffs and ratios
        scores = {measure.name: measure.score(preds, test) for measure in self._measures}
        # store the result
        self.results.append(ResultTuple(parameter_setting, fold_id, scores))
        return scores


class CrossValidator:
    """A simple approach to Cross Validation.

    The CrossValidator object is used to run cross-validation on a model.
    Results are returned in a CVResults object.

    .. code-block:: python

        import ethicml as em

        train, test = em.train_test_split(em.Compas().load())
        hyperparams = {"C": [1, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6]}

        lr_cv = em.CrossValidator(em.LR, hyperparams, folds=3)

        primary = em.Accuracy()
        fair_measure = em.AbsCV()
        cv_results = lr_cv.run(train, measures=[primary, fair_measure])

    :param model: the class (not an instance) of the model for cross validation
    :param hyperparams: a dictionary where the keys are the names of hyperparameters and the values
        are lists of possible values for the hyperparameters
    :param folds: the number of folds
    :param max_parallel: the maximum number of parallel processes; if set to 0, use the default
        which is the number of available CPUs
    """

    def __init__(
        self,
        model: Type[InAlgorithm],
        hyperparams: Mapping[str, Sequence[Any]],
        folds: int = 3,
        max_parallel: int = 0,
    ):
        self.model = model
        self.hyperparams = hyperparams
        self.folds = folds
        self.max_parallel = max_parallel

        keys, values = zip(*hyperparams.items())
        self.experiments: List[Dict[str, Any]] = [dict(zip(keys, v)) for v in product(*values)]

    def run_async(self, train: DataTuple, measures: Optional[List[Metric]] = None) -> CVResults:
        """Run the cross validation experiments asynchronously.

        :param train: the training data
        :param measures:  (Default: None)
        :returns: CVResults
        """
        compute_scores_and_append = _ResultsAccumulator(measures)
        # instantiate all models
        models = [self.model(**experiment) for experiment in self.experiments]
        # create all folds
        data_folds: List[Tuple[DataTuple, DataTuple]] = list(fold_data(train, folds=self.folds))
        # convert to right format
        pair_folds = [TrainValPair(train_fold, val) for (train_fold, val) in data_folds]
        # run everything in parallel
        all_results = run_in_parallel(
            algos=models,
            data=pair_folds,
            seeds=[0] * len(pair_folds),
            num_jobs=self.max_parallel,
        )

        # finally, iterate over all results, compute scores and store them
        for preds_for_dataset, experiment in zip(all_results, self.experiments):
            for i, (preds, (_, val)) in enumerate(zip(preds_for_dataset, data_folds)):
                compute_scores_and_append(experiment, preds, val, i)
        return CVResults(compute_scores_and_append.results, self.model)

    def run(self, train: DataTuple, measures: Optional[List[Metric]] = None) -> CVResults:
        """Run the cross validation experiments.

        :param train: the training data
        :param measures:  (Default: None)
        :returns: CVResults
        """
        compute_scores_and_append = _ResultsAccumulator(measures)
        for (i, (train_fold, val)), experiment in product(
            enumerate(fold_data(train, folds=self.folds)), self.experiments
        ):
            # instantiate model and run it
            model = self.model(**experiment)
            preds = model.run(train_fold, val)
            scores = compute_scores_and_append(experiment, preds, val, i)
            score_string = ", ".join(f"{k}={v:.4g}" for k, v in scores.items())
            print(f"fold: {i}, model: '{model.name}', {score_string}, completed!")
        return CVResults(compute_scores_and_append.results, self.model)



"""Implementation for Louizos et al Variational Fair Autoencoder."""
# pylint: disable=arguments-differ

from itertools import groupby
from typing import List

import torch
from torch import Tensor, nn

from ethicml.data.dataset import Dataset

from .categorical import Categorical


class Decoder(nn.Module):
    """Decoder for VFAE."""

    def __init__(self, dataset: Dataset, deploy: bool = False):
        super().__init__()
        self._deploy = deploy
        self.features: List[str] = dataset.feature_split["x"]

        latent_dims = 50
        hidden_size = 100

        self.shared_net = nn.Sequential()
        in_features = latent_dims + 1
        # add hidden layers according to the number of units specified in "hidden_sizes"
        for depth, num_units in enumerate([hidden_size]):
            self.shared_net.add_module(f"hidden_layer_{depth:d}", nn.Linear(in_features, num_units))
            self.shared_net.add_module(f"ReLu {depth:d}", nn.ReLU())
            in_features = num_units  # update input size to next layer

        def _add_output_layer(feature_group: List[str]) -> nn.Module:
            n_dims = len(feature_group)
            categorical = n_dims > 1  # feature is categorical if it has more than 1 possible output

            layer: nn.Module
            return (
                Categorical(in_features, n_dims)
                if categorical
                else nn.Sequential(nn.Linear(in_features, n_dims))
            )

        self.grouped_features = [
            list(group) for key, group in groupby(self.features, lambda x: x.split("_")[0])
        ]
        self.output_layers = nn.ModuleList(
            [_add_output_layer(feature) for feature in self.grouped_features]
        )

    def forward(self, x: Tensor, s: Tensor) -> Tensor:
        """Forward pass.

        :param x:
        :param s:
        """
        batch_size = x.size(0)
        decoded = self.shared_net(torch.cat((x, s.view(-1, 1)), 1))
        decoded = torch.cat(
            [layer(decoded).view(batch_size, -1) for layer in self.output_layers], dim=1
        )
        return decoded
